# This file is part of allegedb, an object-relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""WindowDict, the core data structure used by allegedb's caching system.

It resembles a dictionary, more specifically a defaultdict-like where retrieving
a key that isn't set will get the highest set key that is lower than the key
you asked for (and thus, keys must be orderable). It is optimized for retrieval
of the same key and neighboring ones repeatedly and in sequence.

"""
from collections import deque, Mapping, MutableMapping, KeysView, ItemsView, ValuesView
from operator import itemgetter, lt, le
from itertools import chain
try:
    import cython
except ImportError:
    class cython:
        def locals(**kwargs):
            def passthru(fun):
                return fun
            return passthru
        cfunc = locals
        int = None
        bint = None

get0 = itemgetter(0)
get1 = itemgetter(1)

# TODO: cancel changes that would put something back to where it was at the start
# This will complicate the update_window functions though, and I don't think it'll
# improve much apart from a bit of efficiency in that the deltas are smaller
# sometimes.
def update_window(turn_from, tick_from, turn_to, tick_to, updfun, branchd):
    """Iterate over a window of time in ``branchd`` and call ``updfun`` on the values"""
    if turn_from in branchd:
        # Not including the exact tick you started from because deltas are *changes*
        for past_state in branchd[turn_from][tick_from+1:]:
            updfun(*past_state)
    for midturn in range(turn_from+1, turn_to):
        if midturn in branchd:
            for past_state in branchd[midturn][:]:
                updfun(*past_state)
    if turn_to in branchd:
        for past_state in branchd[turn_to][:tick_to+1]:
            updfun(*past_state)


def update_backward_window(turn_from, tick_from, turn_to, tick_to, updfun, branchd):
    """Iterate backward over a window of time in ``branchd`` and call ``updfun`` on the values"""
    if turn_from in branchd:
        for future_state in reversed(branchd[turn_from][:tick_from]):
            updfun(*future_state)
    for midturn in range(turn_from-1, turn_to, -1):
        if midturn in branchd:
            for future_state in reversed(branchd[midturn][:]):
                updfun(*future_state)
    if turn_to in branchd:
        for future_state in reversed(branchd[turn_to][tick_to+1:]):
            updfun(*future_state)


class HistoryError(KeyError):
    """You tried to access the past in a bad way."""

    def __init__(self, *args, deleted=False):
        super().__init__(*args)
        self.deleted = deleted


def within_history(rev, windowdict):
    """Return whether the windowdict has history at the revision."""
    if not windowdict:
        return False
    begin = windowdict._past[0][0] if windowdict._past else \
            windowdict._future[-1][0]
    end = windowdict._future[0][0] if windowdict._future else \
          windowdict._past[-1][0]
    return begin <= rev <= end


class WindowDictKeysView(KeysView):
    """Look through all the keys a WindowDict contains."""
    def __contains__(self, rev):
        return rev in self._mapping._keys

    def __iter__(self):
        past = self._mapping._past
        future = self._mapping._future
        if past:
            yield from map(get0, past)
        if future:
            yield from map(get0, reversed(future))


class WindowDictItemsView(ItemsView):
    """Look through everything a WindowDict contains."""
    def __contains__(self, item):
        (rev, v) = item
        mapp = self._mapping
        if not within_history(rev, mapp):
            return False
        for mrev, mv in mapp._past:
            if mrev == rev:
                return mv == v
        for mrev, mv in mapp._future:
            if mrev == rev:
                return mv == v
        return False

    def __iter__(self):
        past = self._mapping._past
        future = self._mapping._future
        if past:
            yield from past
        if future:
            yield from future


class WindowDictPastFutureKeysView(KeysView):
    """View on a WindowDict's keys relative to last lookup"""
    def __iter__(self):
        if not self._mapping.stack:
            return
        yield from map(get0, reversed(self._mapping.stack))

    def __contains__(self, item):
        return item in self._mapping and item in map(get0, self._mapping.stack)


class WindowDictPastFutureItemsView(ItemsView):
    def __iter__(self):
        if not self._mapping.stack:
            return
        yield from reversed(self._mapping.stack)

    def __contains__(self, item):
        if self._out_of_range(item, self._mapping.stack):
            return False
        i0, i1 = item
        if i0 not in self._mapping:
            return False
        for j0, j1 in self._mapping.stack:
            if i0 == j0:
                return i1 == j1
        return False


class WindowDictPastItemsView(WindowDictPastFutureItemsView):
    @staticmethod
    def _out_of_range(item, stack):
        return item[0] < stack[0][0] or item[0] > stack[-1][0]


class WindowDictFutureItemsView(WindowDictPastFutureItemsView):
    """View on a WindowDict's future items relative to last lookup"""
    @staticmethod
    def _out_of_range(item, stack):
        return item[0] < stack[-1][0] or item[0] > stack[0][0]


class WindowDictPastFutureValuesView(ValuesView):
    """Abstract class for views on the past or future values of a WindowDict"""
    def __iter__(self):
        stack = self._mapping.stack
        if not stack:
            return
        yield from map(get1, reversed(stack))

    def __contains__(self, item):
        stack = self._mapping.stack
        if not stack:
            return False
        return item in map(get1, stack)


class WindowDictValuesView(ValuesView):
    """Look through all the values that a WindowDict contains."""
    def __contains__(self, value):
        past = self._mapping._past
        future = self._mapping._future
        if past:
            for rev, v in past:
                if v == value:
                    return True
        if future:
            for rev, v in future:
                if v == value:
                    return True
        return False

    def __iter__(self):
        past = self._mapping._past
        future = self._mapping._future
        if past:
            yield from map(get1, past)
        if future:
            yield from map(get1, future)


class WindowDictPastFutureView(Mapping):
    """Abstract class for historical views on WindowDict"""
    __slots__ = ('stack',)

    def __init__(self, stack):
        self.stack = stack

    def __len__(self):
        stack = self.stack
        if not stack:
            return 0
        return len(stack)


class WindowDictPastView(WindowDictPastFutureView):
    """Read-only mapping of just the past of a WindowDict"""
    def __iter__(self):
        stack = self.stack
        if not stack:
            return
        yield from map(get0, reversed(stack))

    def __getitem__(self, key):
        stack = self.stack
        if not stack or key < stack[0][0] or key > stack[-1][0]:
            raise KeyError
        for rev, value in stack:
            if rev == key:
                return value
        raise KeyError

    def keys(self):
        return WindowDictPastFutureKeysView(self)

    def items(self):
        return WindowDictPastItemsView(self)

    def values(self):
        return WindowDictPastFutureValuesView(self)


class WindowDictFutureView(WindowDictPastFutureView):
    """Read-only mapping of just the future of a WindowDict"""
    def __iter__(self):
        stack = self.stack
        if not stack:
            return
        yield from map(get0, reversed(stack))

    def __getitem__(self, key):
        stack = self.stack
        if not stack or key < stack[-1][0] or key > stack[0][0]:
            raise KeyError
        for rev, value in stack:
            if rev == key:
                return value
        raise KeyError

    def keys(self):
        return WindowDictPastFutureKeysView(self)

    def items(self):
        return WindowDictFutureItemsView(self)

    def values(self):
        return WindowDictPastFutureValuesView(self)


class WindowDictSlice:
    """A slice of history in which the start is earlier than the stop"""
    __slots__ = ['dict', 'slice']

    def __init__(self, dict, slice):
        self.dict = dict
        self.slice = slice

    def __reversed__(self):
        return iter(WindowDictReverseSlice(self.dict, self.slice))

    def __iter__(self):
        dic = self.dict
        if not dic:
            return
        slic = self.slice
        if slic.step is not None:
            for i in range(slic.start or dic.beginning, slic.stop or dic.end+1, slic.step):
                yield dic[i]
            return
        if slic.start is None and slic.stop is None:
            yield from map(get1, dic._past)
            yield from map(get1, reversed(dic._future))
        elif None not in (slic.start, slic.stop):
            if slic.stop == slic.start:
                yield dic[slic.stop]
                return
            past = dic._past
            future = dic._future
            if slic.start < slic.stop:
                left, right = slic.start, slic.stop
                dic.seek(right)
                if not past:
                    return
                if past[-1][0] == right:
                    future.append(past.pop())
                cmp = lt
            else:
                left, right = slic.stop, slic.start
                dic.seek(right)
                if not past:
                    return
                cmp = le
            it = iter(past)
            p0, p1 = next(it)
            while cmp(p0, left):
                p0, p1 = next(it)
            else:
                yield p1
            yield from map(get1, it)
        elif slic.start is None:
            stac = dic._past + list(reversed(dic._future))
            while stac and stac[-1][0] >= slic.stop:
                stac.pop()
            yield from map(get1, stac)
            return
        else:  # slic.stop is None
            if not dic._past and not dic._future:
                return
            chan = chain(dic._past, reversed(dic._future))
            nxt = next(chan)
            while nxt[0] < slic.start:
                try:
                    nxt = next(chan)
                except StopIteration:
                    return
            yield get1(nxt)
            yield from map(get1, chan)


class WindowDictReverseSlice:
    """A slice of history in which the start is later than the stop"""
    __slots__ = ['dict', 'slice']

    def __init__(self, dict, slice):
        self.dict = dict
        self.slice = slice

    def __reversed__(self):
        return iter(WindowDictSlice(self.dict, self.slice))

    def __iter__(self):
        dic = self.dict
        if not dic:
            return
        slic = self.slice
        if slic.step is not None:
            for i in range(slic.start or dic.end, slic.stop or dic.beginning, slic.step):
                yield dic[i]
            return
        if slic.start is None and slic.stop is None:
            yield from map(get1, dic._future)
            yield from map(get1, reversed(dic._past))
        elif None not in (slic.start, slic.stop):
            if slic.start == slic.stop:
                yield dic[slic.stop]
                return
            if slic.start < slic.stop:
                left, right = slic.start, slic.stop
                dic.seek(right)
                it = reversed(dic._past)
                next(it)
                cmp = lt
            else:
                left, right = slic.stop, slic.start
                dic.seek(right)
                it = reversed(dic._past)
                cmp = le
            for frev, fv in it:
                if cmp(frev, left):
                    return
                yield fv
        elif slic.start is None:
            stac = dic._past + list(reversed(dic._future))
            while stac and stac[-1][0] >= slic.stop:
                stac.pop()
            yield from map(get1, reversed(stac))
        else:  # slic.stop is None
            stac = deque(dic._past)
            stac.extend(reversed(dic._future))
            while stac and stac[0][0] < slic.start:
                stac.popleft()
            yield from map(get1, reversed(stac))


class WindowDict(MutableMapping):
    """A dict that keeps every value that a variable has had over time.

    Look up a revision number in this dict and it will give you the
    effective value as of that revision. Keys should always be
    revision numbers.

    Optimized for the cases where you look up the same revision
    repeatedly, or its neighbors.

    This supports slice notation to get all values in a given
    time-frame. If you do not supply a step, you'll just get the
    values, with no indication of when they're from exactly --
    so explicitly supply a step of 1 to get the value at each point in
    the slice, or use the ``future`` and ``past`` methods to get read-only
    mappings of data relative to a particular revision.

    Unlike slices of eg. lists, you can slice with a start greater than the stop
    even if you don't supply a step. That will get you values in reverse order.

    """
    __slots__ = ('_future', '_past', '_keys', '_beginning')

    def future(self, rev=None):
        """Return a Mapping of items after the given revision.

        Default revision is the last one looked up.

        """
        if rev is not None:
            self.seek(rev)
        return WindowDictFutureView(self._future)

    def past(self, rev=None):
        """Return a Mapping of items at or before the given revision.

        Default revision is the last one looked up.

        """
        if rev is not None:
            self.seek(rev)
        return WindowDictPastView(self._past)

    @cython.locals(rev=cython.int, past_end=cython.int, future_start=cython.int)
    def seek(self, rev):
        """Arrange the caches to help look up the given revision."""
        # TODO: binary search? Perhaps only when one or the other
        # stack is very large?
        if not self:
            return
        if type(rev) is not int:
            raise TypeError("rev must be int")
        past = self._past
        future = self._future
        if future:
            appender = past.append
            popper = future.pop
            future_start = future[-1][0]
            while future_start <= rev:
                appender(popper())
                if future:
                    future_start = future[-1][0]
                else:
                    break
        if past:
            popper = past.pop
            appender = future.append
            past_end = past[-1][0]
            while past_end > rev:
                appender(popper())
                if past:
                    past_end = past[-1][0]
                else:
                    break

    def rev_gettable(self, rev: int) -> bool:
        beg = self._beginning
        if beg is None:
            return False
        return rev >= beg

    def rev_before(self, rev: int) -> int:
        """Return the latest past rev on which the value changed."""
        self.seek(rev)
        if self._past:
            return self._past[-1][0]

    def rev_after(self, rev: int) -> int:
        """Return the earliest future rev on which the value will change."""
        self.seek(rev)
        if self._future:
            return self._future[-1][0]

    def truncate(self, rev: int) -> None:
        """Delete everything after the given revision."""
        self.seek(rev)
        self._keys.difference_update(map(get0, self._future))
        self._future = []
        if not self._past:
            self._beginning = None

    @property
    def end(self) -> int:
        if self._future:
            return self._future[0][0]
        elif self._past:
            return self._past[-1][0]
        else:
            raise HistoryError("No history yet")

    def keys(self):
        return WindowDictKeysView(self)

    def items(self):
        return WindowDictItemsView(self)

    def values(self):
        return WindowDictValuesView(self)

    def __bool__(self):
        return bool(self._past) or bool(self._future)

    def __init__(self, data=None):
        if not data:
            self._past = []
        elif hasattr(data, 'items'):
            self._past = list(data.items())
        else:
            # assume it's an orderable sequence of pairs
            self._past = list(data)
        self._past.sort()
        self._future = []
        self._keys = set(map(get0, self._past))
        self._beginning = None if not self._past else self._past[0][0]

    def __iter__(self):
        if not self:
            return
        if self._past:
            yield from map(get0, self._past)
        if self._future:
            yield from map(get0, self._future)

    def __contains__(self, item):
        return item in self._keys

    def __len__(self):
        return len(self._past) + len(self._future)

    def __getitem__(self, rev):
        if not self:
            raise HistoryError("No history yet")
        if isinstance(rev, slice):
            if None not in (rev.start, rev.stop) and rev.start > rev.stop:
                return WindowDictReverseSlice(self, rev)
            return WindowDictSlice(self, rev)
        self.seek(rev)
        past = self._past
        if not past:
            raise HistoryError(
                "Revision {} is before the start of history".format(rev)
            )
        return past[-1][1]

    @cython.locals(past_end=cython.int, rev=cython.int)
    def __setitem__(self, rev, v):
        if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap'):
            v = v.unwrap()
        past = self._past
        if past or self._future:
            self.seek(rev)
            past_end = -1 if not past else past[-1][0]
            if not past:
                past.append((rev, v))
                self._beginning = rev
            elif past_end == rev:
                past[-1] = (rev, v)
            else:
                past.append((rev, v))
        else:
            past.append((rev, v))
            self._beginning = rev
        self._keys.add(rev)

    @property
    def beginning(self):
        if self._beginning is None:
            raise HistoryError("No history yet")
        return self._beginning

    @cython.locals(rev=cython.int, past_end=cython.int)
    def __delitem__(self, rev):
        # Not checking for rev's presence at the beginning because
        # to do so would likely require iterating thru history,
        # which I have to do anyway in deleting.
        # But handle degenerate case.
        if not self:
            raise HistoryError("Tried to delete from an empty WindowDict")
        if not self.beginning <= rev <= self.end:
            raise HistoryError("Rev outside of history: {}".format(rev))
        self.seek(rev)
        past = self._past
        past_end = -1 if not past else past[-1][0]
        if not past or past_end != rev:
            raise HistoryError("Rev not present: {}".format(rev))
        del past[-1]
        if not past:
            if self._future:
                self._beginning = self._future[-1][0]
            else:
                self._beginning = None
        self._keys.remove(rev)

    def __repr__(self):
        me = dict(self._past)
        me.update(self._future)
        return "{}({})".format(self.__class__.__name__, me)


class FuturistWindowDict(WindowDict):
    """A WindowDict that does not let you rewrite the past."""
    __slots__ = ('_future', '_past', '_beginning')

    def __setitem__(self, rev, v):
        if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap'):
            v = v.unwrap()
        self.seek(rev)
        past = self._past
        future = self._future
        if future:
            raise HistoryError(
                "Already have some history after {}".format(rev)
            )
        if not past:
            self._beginning = rev
            past.append((rev, v))
        elif rev > past[-1][0]:
            past.append((rev, v))
        elif rev == past[-1][0]:
            past[-1] = (rev, v)
        else:
            raise HistoryError(
                "Already have some history after {} "
                "(and my seek function is broken?)".format(rev)
            )
        self._keys.add(rev)


class TurnDict(FuturistWindowDict):
    __slots__ = ('_future', '_past')
    cls = FuturistWindowDict

    def __setitem__(self, turn, value):
        if type(value) is not FuturistWindowDict:
            value = FuturistWindowDict(value)
        FuturistWindowDict.__setitem__(self, turn, value)


class SettingsTurnDict(WindowDict):
    __slots__ = ('_future', '_past')
    cls = WindowDict

    def __setitem__(self, turn, value):
        if type(value) is not WindowDict:
            value = WindowDict(value)
        WindowDict.__setitem__(self, turn, value)



# This file is part of allegedb, an object relational mapper for graphs.
# Copyright (c) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Wrapper to run SQL queries in a lightly abstracted way, such that
code that's more to do with the queries than with the data per se
doesn't pollute the other files so much.

"""
import os
from collections import MutableMapping
from sqlite3 import IntegrityError as sqliteIntegError
try:
    # python 2
    import wrap
except ImportError:
    # python 3
    from allegedb import wrap
wrappath = os.path.dirname(wrap.__file__)
alchemyIntegError = None
try:
    from sqlalchemy.exc import IntegrityError as alchemyIntegError
except ImportError:
    pass


IntegrityError = (
    alchemyIntegError, sqliteIntegError
) if alchemyIntegError is not None else sqliteIntegError


class TimeError(ValueError):
    """Exception class for problems with the time model"""


class GlobalKeyValueStore(MutableMapping):
    """A dict-like object that keeps its contents in a table.

    Mostly this is for holding the current branch and revision.

    """
    def __init__(self, qe):
        self.qe = qe

    def __iter__(self):
        if hasattr(self.qe, '_global_cache'):
            yield from self.qe._global_cache
            return
        for (k, v) in self.qe.global_items():
            yield k

    def __len__(self):
        if hasattr(self.qe, '_global_cache'):
            return len(self.qe._global_cache)
        return self.qe.ctglobal()

    def __getitem__(self, k):
        if hasattr(self.qe, '_global_cache'):
            return self.qe._global_cache[k]
        return self.qe.global_get(k)

    def __setitem__(self, k, v):
        self.qe.global_set(k, v)
        if hasattr(self.qe, '_global_cache'):
            self.qe._global_cache[k] = v

    def __delitem__(self, k):
        self.qe.global_del(k)
        if hasattr(self.qe, '_global_cache'):
            del self.qe._global_cache[k]


class QueryEngine(object):
    """Wrapper around either a DBAPI2.0 connection or an
    Alchemist. Provides methods to run queries using either.

    """
    path = os.path.dirname(__file__)

    def __init__(
            self, dbstring, connect_args, alchemy,
            pack=None, unpack=None
    ):
        """If ``alchemy`` is True and ``dbstring`` is a legit database URI,
        instantiate an Alchemist and start a transaction with
        it. Otherwise use sqlite3.

        You may pass an already created sqlalchemy :class:`Engine`
        object in place of ``dbstring`` if you wish. I'll still create
        my own transaction though.

        """
        dbstring = dbstring or 'sqlite:///:memory:'

        def alchem_init(dbstring, connect_args):
            from sqlalchemy import create_engine
            from sqlalchemy.engine.base import Engine
            from allegedb.alchemy import Alchemist
            if isinstance(dbstring, Engine):
                self.engine = dbstring
            else:
                self.engine = create_engine(
                    dbstring,
                    connect_args=connect_args
                )
            self.alchemist = Alchemist(self.engine)
            self.transaction = self.alchemist.conn.begin()

        def lite_init(dbstring, connect_args):
            from sqlite3 import connect, Connection
            from json import load
            self.strings = load(
                open(os.path.join(self.path, 'sqlite.json'))
            )
            if isinstance(dbstring, Connection):
                self.connection = dbstring
            else:
                if dbstring.startswith('sqlite:'):
                    slashidx = dbstring.rindex('/')
                    dbstring = dbstring[slashidx+1:]
                self.connection = connect(dbstring)

        if alchemy:
            try:
                alchem_init(dbstring, connect_args)
            except ImportError:
                lite_init(dbstring, connect_args)
        else:
            lite_init(dbstring, connect_args)

        self.globl = GlobalKeyValueStore(self)
        self._branches = {}
        self._nodevals2set = []
        self._edgevals2set = []
        self._graphvals2set = []
        self._nodes2set = []
        self._edges2set = []
        self._btts = set()
        if unpack is None:
            from ast import literal_eval as unpack
        self.pack = pack or repr
        self.unpack = unpack

    def sql(self, stringname, *args, **kwargs):
        """Wrapper for the various prewritten or compiled SQL calls.

        First argument is the name of the query, either a key in
        ``sqlite.json`` or a method name in
        ``allegedb.alchemy.Alchemist``. The rest of the arguments are
        parameters to the query.

        """
        if hasattr(self, 'alchemist'):
            return getattr(self.alchemist, stringname)(*args, **kwargs)
        else:
            s = self.strings[stringname]
            return self.connection.cursor().execute(
                s.format(**kwargs) if kwargs else s, args
            )

    def sqlmany(self, stringname, *args):
        """Wrapper for executing many SQL calls on my connection.

        First arg is the name of a query, either a key in the
        precompiled JSON or a method name in
        ``allegedb.alchemy.Alchemist``. Remaining arguments should be
        tuples of argument sequences to be passed to the query.

        """
        if hasattr(self, 'alchemist'):
            return getattr(self.alchemist.many, stringname)(*args)
        s = self.strings[stringname]
        return self.connection.cursor().executemany(s, args)

    def have_graph(self, graph):
        """Return whether I have a graph by this name."""
        graph = self.pack(graph)
        return bool(self.sql('graphs_named', graph).fetchone()[0])

    def new_graph(self, graph, typ):
        """Declare a new graph by this name of this type."""
        graph = self.pack(graph)
        return self.sql('new_graph', graph, typ)

    def del_graph(self, graph):
        """Delete all records to do with the graph"""
        g = self.pack(graph)
        self.sql('del_edge_val_graph', g)
        self.sql('del_node_val_graph', g)
        self.sql('del_edge_val_graph', g)
        self.sql('del_edges_graph', g)
        self.sql('del_nodes_graph', g)
        self.sql('del_graph', g)

    def graph_type(self, graph):
        """What type of graph is this?"""
        graph = self.pack(graph)
        return self.sql('graph_type', graph).fetchone()[0]

    def have_branch(self, branch):
        """Return whether the branch thus named exists in the database."""
        return bool(self.sql('ctbranch', branch).fetchone()[0])

    def all_branches(self):
        """Return all the branch data in tuples of (branch, parent,
        parent_turn).

        """
        return self.sql('branches_dump').fetchall()

    def global_get(self, key):
        """Return the value for the given key in the ``globals`` table."""
        key = self.pack(key)
        r = self.sql('global_get', key).fetchone()
        if r is None:
            raise KeyError("Not set")
        return self.unpack(r[0])

    def global_items(self):
        """Iterate over (key, value) pairs in the ``globals`` table."""
        for (k, v) in self.sql('global_dump'):
            yield (self.unpack(k), self.unpack(v))

    def global_set(self, key, value):
        """Set ``key`` to ``value`` globally (not at any particular branch or
        revision)

        """
        (key, value) = map(self.pack, (key, value))
        try:
            return self.sql('global_insert', key, value)
        except IntegrityError:
            return self.sql('global_update', value, key)

    def global_del(self, key):
        """Delete the global record for the key."""
        key = self.pack(key)
        return self.sql('global_del', key)

    def new_branch(self, branch, parent, parent_turn, parent_tick):
        """Declare that the ``branch`` is descended from ``parent`` at
        ``parent_turn``, ``parent_tick``

        """
        return self.sql('branches_insert', branch, parent, parent_turn, parent_tick, parent_turn, parent_tick)

    def update_branch(self, branch, parent, parent_turn, parent_tick, end_turn, end_tick):
        return self.sql('update_branches', parent, parent_turn, parent_tick, end_turn, end_tick, branch)

    def set_branch(self, branch, parent, parent_turn, parent_tick, end_turn, end_tick):
        try:
            self.sql('branches_insert', branch, parent, parent_turn, parent_tick, end_turn, end_tick)
        except IntegrityError:
            self.update_branch(branch, parent, parent_turn, parent_tick, end_turn, end_tick)

    def new_turn(self, branch, turn, end_tick=0, plan_end_tick=0):
        return self.sql('turns_insert', branch, turn, end_tick, plan_end_tick)

    def update_turn(self, branch, turn, end_tick, plan_end_tick):
        return self.sql('update_turns', end_tick, plan_end_tick, branch, turn)

    def set_turn(self, branch, turn, end_tick, plan_end_tick):
        try:
            return self.sql('turns_insert', branch, turn, end_tick, plan_end_tick)
        except IntegrityError:
            return self.sql('update_turns', end_tick, plan_end_tick, branch, turn)

    def turns_dump(self):
        return self.sql('turns_dump')

    def graph_val_dump(self):
        """Yield the entire contents of the graph_val table."""
        self._flush_graph_val()
        for (graph, key, branch, turn, tick, value) in self.sql('graph_val_dump'):
            yield (
                self.unpack(graph),
                self.unpack(key),
                branch,
                turn,
                tick,
                self.unpack(value)
            )

    def _flush_graph_val(self):
        """Send all new and changed graph values to the database."""
        if not self._graphvals2set:
            return
        delafter = {}
        for graph, key, branch, turn, tick, value in self._graphvals2set:
            if (graph, key, branch) in delafter:
                delafter[graph, key, branch] = min((
                    (turn, tick),
                    delafter[graph, key, branch]
                ))
            else:
                delafter[graph, key, branch] = (turn, tick)
        self.sqlmany(
            'del_graph_val_after',
            *((graph, key, branch, turn, turn, tick)
              for ((graph, key, branch), (turn, tick)) in delafter.items())
        )
        self.sqlmany('graph_val_insert', *self._graphvals2set)
        self._graphvals2set = []

    def graph_val_set(self, graph, key, branch, turn, tick, value):
        if (branch, turn, tick) in self._btts:
            raise TimeError
        self._btts.add((branch, turn, tick))
        graph, key, value = map(self.pack, (graph, key, value))
        self._graphvals2set.append((graph, key, branch, turn, tick, value))

    def graph_val_del_time(self, branch, turn, tick):
        self._flush_graph_val()
        self.sql('graph_val_del_time', branch, turn, tick)
        self._btts.discard((branch, turn, tick))

    def graphs_types(self):
        for (graph, typ) in self.sql('graphs_types'):
            yield (self.unpack(graph), typ)

    def _flush_nodes(self):
        if not self._nodes2set:
            return
        # delete history that is to be overwritten due to paradox
        cleanups = {}
        for graph, node, branch, turn, tick, extant in self._nodes2set:
            if (graph, node, branch) in cleanups:
                cleanups[graph, node, branch] = min((
                    (turn, tick), cleanups[graph, node, branch]
                ))
            else:
                cleanups[graph, node, branch] = (turn, tick)
        self.sqlmany('del_nodes_after', *(k + (turn, turn, tick) for k, (turn, tick) in cleanups.items()))
        self.sqlmany('nodes_insert', *self._nodes2set)
        self._nodes2set = []

    def exist_node(self, graph, node, branch, turn, tick, extant):
        """Declare that the node exists or doesn't.

        Inserts a new record or updates an old one, as needed.

        """
        if (branch, turn, tick) in self._btts:
            raise TimeError
        self._btts.add((branch, turn, tick))
        self._nodes2set.append((self.pack(graph), self.pack(node), branch, turn, tick, extant))

    def nodes_del_time(self, branch, turn, tick):
        self._flush_nodes()
        self.sql('nodes_del_time', branch, turn, tick)
        self._btts.discard((branch, turn, tick))

    def nodes_dump(self):
        """Dump the entire contents of the nodes table."""
        self._flush_nodes()
        for (graph, node, branch, turn,tick, extant) in self.sql('nodes_dump'):
            yield (
                self.unpack(graph),
                self.unpack(node),
                branch,
                turn,
                tick,
                bool(extant)
            )

    def node_val_dump(self):
        """Yield the entire contents of the node_val table."""
        self._flush_node_val()
        for (
                graph, node, key, branch, turn, tick, value
        ) in self.sql('node_val_dump'):
            yield (
                self.unpack(graph),
                self.unpack(node),
                self.unpack(key),
                branch,
                turn,
                tick,
                self.unpack(value)
            )

    def _flush_node_val(self):
        if not self._nodevals2set:
            return
        delafter = {}
        for graph, node, key, branch, turn, tick, value in self._nodevals2set:
            if (graph, node, key, branch) in delafter:
                delafter[graph, node, key, branch] = min((
                    (turn, tick),
                    delafter[graph, node, key, branch]
                ))
            else:
                delafter[graph, node, key, branch] = (turn, tick)
        if delafter:
            self.sqlmany(
                'del_node_val_after',
                *((graph, node, key, branch, turn, turn, tick)
                  for ((graph, node, key, branch), (turn, tick)) in
                  delafter.items())
            )
        self.sqlmany('node_val_insert', *self._nodevals2set)
        self._nodevals2set = []

    def node_val_set(self, graph, node, key, branch, turn, tick, value):
        """Set a key-value pair on a node at a specific branch and revision"""
        if (branch, turn, tick) in self._btts:
            raise TimeError
        self._btts.add((branch, turn, tick))
        graph, node, key, value = map(self.pack, (graph, node, key, value))
        self._nodevals2set.append((graph, node, key, branch, turn, tick, value))

    def node_val_del_time(self, branch, turn, tick):
        self._flush_node_val()
        self.sql('node_val_del_time', branch, turn, tick)
        self._btts.discard((branch, turn, tick))

    def edges_dump(self):
        """Dump the entire contents of the edges table."""
        self._flush_edges()
        for (
                graph, orig, dest, idx, branch, turn, tick, extant
        ) in self.sql('edges_dump'):
            yield (
                self.unpack(graph),
                self.unpack(orig),
                self.unpack(dest),
                idx,
                branch,
                turn,
                tick,
                bool(extant)
            )

    def _flush_edges(self):
        if not self._edges2set:
            return
        delafter = {}
        for graph, orig, dest, idx, branch, turn, tick, extant in self._edges2set:
            key = graph, orig, dest, idx, branch
            if key in delafter:
                delafter[key] = min((
                    (turn, tick),
                    delafter[key]
                ))
            else:
                delafter[key] = (turn, tick)
        if delafter:
            self.sqlmany(
                'del_edges_after',
                *((graph, orig, dest, idx, branch, turn, turn, tick)
                  for ((graph, orig, dest, idx, branch), (turn, tick)) in
                  delafter.items())
            )
        self.sqlmany('edges_insert', *self._edges2set)
        self._edges2set = []

    def exist_edge(self, graph, orig, dest, idx, branch, turn, tick, extant):
        """Declare whether or not this edge exists."""
        if (branch, turn, tick) in self._btts:
            raise TimeError
        self._btts.add((branch, turn, tick))
        graph, orig, dest = map(self.pack, (graph, orig, dest))
        self._edges2set.append((graph, orig, dest, idx, branch, turn, tick, extant))

    def edges_del_time(self, branch, turn, tick):
        self._flush_edges()
        self.sql('edges_del_time', branch, turn, tick)
        self._btts.discard((branch, turn, tick))

    def edge_val_dump(self):
        """Yield the entire contents of the edge_val table."""
        self._flush_edge_val()
        for (
                graph, orig, dest, idx, key, branch, turn, tick, value
        ) in self.sql('edge_val_dump'):
            yield (
                self.unpack(graph),
                self.unpack(orig),
                self.unpack(dest),
                idx,
                self.unpack(key),
                branch,
                turn,
                tick,
                self.unpack(value)
            )

    def _flush_edge_val(self):
        if not self._edgevals2set:
            return
        delafter = {}
        for graph, orig, dest, idx, key, branch, turn, tick, value in self._edgevals2set:
            dkey = graph, orig, dest, idx, key, branch
            if dkey in delafter:
                delafter[dkey] = min((
                    (turn, tick), delafter[dkey]
                ))
            else:
                delafter[dkey] = (turn, tick)
        self.sqlmany(
            'del_edge_val_after',
            *((graph, orig, dest, idx, key, branch, turn, turn, tick)
              for ((graph, orig, dest, idx, key, branch), (turn, tick))
              in delafter.items())
        )
        self.sqlmany('edge_val_insert', *self._edgevals2set)
        self._edgevals2set = []

    def edge_val_set(self, graph, orig, dest, idx, key, branch, turn, tick, value):
        """Set this key of this edge to this value."""
        if (branch, turn, tick) in self._btts:
            raise TimeError
        self._btts.add((branch, turn, tick))
        graph, orig, dest, key, value = map(self.pack, (graph, orig, dest, key, value))
        self._edgevals2set.append(
            (graph, orig, dest, idx, key, branch, turn, tick, value)
        )

    def edge_val_del_time(self, branch, turn, tick):
        self._flush_edge_val()
        self.sql('edge_val_del_time', branch, turn, tick)
        self._btts.discard((branch, turn, tick))

    def plans_dump(self):
        return self.sql('plans_dump')

    def plans_insert(self, plan_id, branch, turn, tick):
        return self.sql('plans_insert', plan_id, branch, turn, tick)

    def plans_insert_many(self, many):
        return self.sqlmany('plans_insert', *many)

    def plan_ticks_insert(self, plan_id, turn, tick):
        return self.sql('plan_ticks_insert', plan_id, turn, tick)

    def plan_ticks_insert_many(self, many):
        return self.sqlmany('plan_ticks_insert', *many)

    def plan_ticks_dump(self):
        return self.sql('plan_ticks_dump')

    def initdb(self):
        """Create tables and indices as needed."""
        if hasattr(self, 'alchemist'):
            self.alchemist.meta.create_all(self.engine)
            if 'branch' not in self.globl:
                self.globl['branch'] = 'trunk'
            if 'rev' not in self.globl:
                self.globl['rev'] = 0
            return
        from sqlite3 import OperationalError
        cursor = self.connection.cursor()
        try:
            cursor.execute('SELECT * FROM global;')
        except OperationalError:
            cursor.execute(self.strings['create_global'])
        if 'branch' not in self.globl:
            self.globl['branch'] = 'trunk'
        if 'turn' not in self.globl:
            self.globl['turn'] = 0
        if 'tick' not in self.globl:
            self.globl['tick'] = 0
        for table in (
            'branches',
            'turns',
            'graphs',
            'graph_val',
            'nodes',
            'node_val',
            'edges',
            'edge_val',
            'plans',
            'plan_ticks'
        ):
            try:
                cursor.execute('SELECT * FROM ' + table + ';')
            except OperationalError:
                cursor.execute(self.strings['create_' + table])

    def flush(self):
        """Put all pending changes into the SQL transaction."""
        self._flush_nodes()
        self._flush_edges()
        self._flush_graph_val()
        self._flush_node_val()
        self._flush_edge_val()

    def commit(self):
        """Commit the transaction"""
        self.flush()
        if hasattr(self, 'transaction') and self.transaction.is_active:
            self.transaction.commit()
        elif hasattr(self, 'connection'):
            self.connection.commit()

    def close(self):
        """Commit the transaction, then close the connection"""
        self.commit()
        if hasattr(self, 'connection'):
            self.connection.close()



# This file is part of allegedb, an object relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""SQLAlchemy code to generate the SQL used by the allegedb ORM

If SQLAlchemy is installed at runtime, this will use it to generate SQL on the fly;
if it is not, allegedb can use a pregenerated file "sqlite.json" to store graphs in
a SQLite database. Generate sqlite.json by running this on the command line:

python3 alchemy.py >sqlite.json

"""
from functools import partial
from sqlalchemy import (
    Table,
    Index,
    Column,
    CheckConstraint,
    ForeignKeyConstraint,
    INT,
    TEXT,
    BOOLEAN,
    MetaData,
    ForeignKey,
    select,
    func,
)


BaseColumn = Column
Column = partial(BaseColumn, nullable=False)


from sqlalchemy.sql import bindparam, and_, or_
from json import dumps
from functools import partial


def tables_for_meta(meta):
    Table(
        'global', meta,
        Column('key', TEXT, primary_key=True),
        Column('value', TEXT, nullable=True)
    )
    Table(
        'branches', meta,
        Column(
            'branch', TEXT, ForeignKey('branches.parent'),
            primary_key=True, default='trunk'
        ),
        Column('parent', TEXT, default='trunk', nullable=True),
        Column('parent_turn', INT, default=0),
        Column('parent_tick', INT, default=0),
        Column('end_turn', INT, default=0),
        Column('end_tick', INT, default=0),
        CheckConstraint('branch<>parent')
    )
    Table(
        'turns', meta,
        Column('branch', TEXT, primary_key=True),
        Column('turn', INT, primary_key=True),
        Column('end_tick', INT),
        Column('plan_end_tick', INT)
    )
    Table(
        'graphs', meta,
        Column('graph', TEXT, primary_key=True),
        Column('type', TEXT, default='Graph'),
        CheckConstraint(
            "type IN ('Graph', 'DiGraph', 'MultiGraph', 'MultiDiGraph')"
        )
    )
    Table(
        'graph_val', meta,
        Column('graph', TEXT, ForeignKey('graphs.graph'),
               primary_key=True),
        Column('key', TEXT, primary_key=True),
        Column('branch', TEXT, ForeignKey('branches.branch'),
               primary_key=True, default='trunk'),
        Column('turn', INT, primary_key=True, default=0),
        Column('tick', INT, primary_key=True, default=0),
        Column('value', TEXT, nullable=True)
    )
    Table(
        'nodes', meta,
        Column('graph', TEXT, ForeignKey('graphs.graph'),
               primary_key=True),
        Column('node', TEXT, primary_key=True),
        Column('branch', TEXT, ForeignKey('branches.branch'),
               primary_key=True, default='trunk'),
        Column('turn', INT, primary_key=True, default=0),
        Column('tick', INT, primary_key=True, default=0),
        Column('extant', BOOLEAN)
    )
    Table(
        'node_val', meta,
        Column('graph', TEXT, primary_key=True),
        Column('node', TEXT, primary_key=True),
        Column('key', TEXT, primary_key=True),
        Column('branch', TEXT, ForeignKey('branches.branch'),
               primary_key=True, default='trunk'),
        Column('turn', INT, primary_key=True, default=0),
        Column('tick', INT, primary_key=True, default=0),
        Column('value', TEXT, nullable=True),
        ForeignKeyConstraint(
            ['graph', 'node'], ['nodes.graph', 'nodes.node']
        )
    )
    Table(
        'edges', meta,
        Column('graph', TEXT, ForeignKey('graphs.graph'),
               primary_key=True),
        Column('orig', TEXT, primary_key=True),
        Column('dest', TEXT, primary_key=True),
        Column('idx', INT, primary_key=True),
        Column('branch', TEXT, ForeignKey('branches.branch'),
               primary_key=True, default='trunk'),
        Column('turn', INT, primary_key=True, default=0),
        Column('tick', INT, primary_key=True, default=0),
        Column('extant', BOOLEAN),
        ForeignKeyConstraint(
            ['graph', 'orig'], ['nodes.graph', 'nodes.node']
        ),
        ForeignKeyConstraint(
            ['graph', 'dest'], ['nodes.graph', 'nodes.node']
        )
    )
    Table(
        'edge_val', meta,
        Column('graph', TEXT, primary_key=True),
        Column('orig', TEXT, primary_key=True),
        Column('dest', TEXT, primary_key=True),
        Column('idx', INT, primary_key=True),
        Column('key', TEXT, primary_key=True),
        Column('branch', TEXT, ForeignKey('branches.branch'),
               primary_key=True, default='trunk'),
        Column('turn', INT, primary_key=True, default=0),
        Column('tick', INT, primary_key=True, default=0),
        Column('value', TEXT, nullable=True),
        ForeignKeyConstraint(
            ['graph', 'orig', 'dest', 'idx'],
            ['edges.graph', 'edges.orig', 'edges.dest', 'edges.idx']
        )
    )
    Table(
        'plans', meta,
        Column('id', INT, primary_key=True),
        Column('branch', TEXT),
        Column('turn', INT),
        Column('tick', INT)
    )
    Table(
        'plan_ticks', meta,
        Column('plan_id', INT, primary_key=True),
        Column('turn', INT, primary_key=True),
        Column('tick', INT, primary_key=True),
        ForeignKeyConstraint(
            ('plan_id',),
            ('plans.id',)
        )
    )
    return meta.tables


def indices_for_table_dict(table):
    return {}


def queries_for_table_dict(table):
    r = {
        'global_get': select(
            [table['global'].c.value]
        ).where(
            table['global'].c.key == bindparam('key')
        ),
        'global_update': table['global'].update().values(
            value=bindparam('value')
        ).where(table['global'].c.key == bindparam('key')),
        'new_graph': table['graphs'].insert().values(
            graph=bindparam('graph'),
            type=bindparam('type')
        ),
        'graph_type': select(
            [table['graphs'].c.type]
        ).where(
            table['graphs'].c.graph == bindparam('graph')
        ),
        'del_edge_val_graph': table['edge_val'].delete().where(
            table['edge_val'].c.graph == bindparam('graph')
        ),
        'del_edge_val_after': table['edge_val'].delete().where(and_(
            table['edge_val'].c.graph == bindparam('graph'),
            table['edge_val'].c.orig == bindparam('orig'),
            table['edge_val'].c.dest == bindparam('dest'),
            table['edge_val'].c.idx == bindparam('idx'),
            table['edge_val'].c.key == bindparam('key'),
            table['edge_val'].c.branch == bindparam('branch'),
            or_(
                table['edge_val'].c.turn > bindparam('turn'),
                and_(
                    table['edge_val'].c.turn == bindparam('turn'),
                    table['edge_val'].c.tick >= bindparam('tick')
                )
            )
        )),
        'del_edges_graph': table['edges'].delete().where(
            table['edges'].c.graph == bindparam('graph')
        ),
        'del_edges_after': table['edges'].delete().where(and_(
            table['edges'].c.graph == bindparam('graph'),
            table['edges'].c.orig == bindparam('orig'),
            table['edges'].c.dest == bindparam('dest'),
            table['edges'].c.idx == bindparam('idx'),
            table['edges'].c.branch == bindparam('branch'),
            or_(
                table['edges'].c.turn > bindparam('turn'),
                and_(
                    table['edges'].c.turn == bindparam('turn'),
                    table['edges'].c.tick >= bindparam('tick')
                )
            )
        )),
        'del_nodes_graph': table['nodes'].delete().where(
            table['nodes'].c.graph == bindparam('graph')
        ),
        'del_nodes_after': table['nodes'].delete().where(and_(
            table['nodes'].c.graph == bindparam('graph'),
            table['nodes'].c.node == bindparam('node'),
            table['nodes'].c.branch == bindparam('branch'),
            or_(
                table['nodes'].c.turn > bindparam('turn'),
                and_(
                    table['nodes'].c.turn == bindparam('turn'),
                    table['nodes'].c.tick >= bindparam('tick')
                )
            )
        )),
        'del_node_val_graph': table['node_val'].delete().where(
            table['node_val'].c.graph == bindparam('graph')
        ),
        'del_node_val_after': table['node_val'].delete().where(and_(
            table['node_val'].c.graph == bindparam('graph'),
            table['node_val'].c.node == bindparam('node'),
            table['node_val'].c.key == bindparam('key'),
            table['node_val'].c.branch == bindparam('branch'),
            or_(
                table['node_val'].c.turn > bindparam('turn'),
                and_(
                    table['node_val'].c.turn == bindparam('turn'),
                    table['node_val'].c.tick >= bindparam('tick')
                )
            )
        )),
        'del_graph': table['graphs'].delete().where(
            table['graphs'].c.graph == bindparam('graph')
        ),
        'del_graph_val_after': table['graph_val'].delete().where(and_(
            table['graph_val'].c.graph == bindparam('graph'),
            table['graph_val'].c.key == bindparam('key'),
            table['graph_val'].c.branch == bindparam('branch'),
            or_(
                table['graph_val'].c.turn > bindparam('turn'),
                and_(
                    table['graph_val'].c.turn == bindparam('turn'),
                    table['graph_val'].c.tick >= bindparam('tick')
                )
            )
        )),
        'global_delete': table['global'].delete().where(
            table['global'].c.key == bindparam('key')
        ),
        'graphs_types': select([
            table['graphs'].c.graph,
            table['graphs'].c.type
        ]),
        'graphs_named': select([func.COUNT()]).select_from(table['graphs']).where(
            table['graphs'].c.graph == bindparam('graph')
        ),
        'update_branches': table['branches'].update().values(
            parent=bindparam('parent'),
            parent_turn=bindparam('parent_turn'),
            parent_tick=bindparam('parent_tick'),
            end_turn=bindparam('end_turn'),
            end_tick=bindparam('end_tick')
        ).where(table['branches'].c.branch == bindparam('branch')),
        'update_turns': table['turns'].update().values(
            end_tick=bindparam('end_tick'),
            plan_end_tick=bindparam('plan_end_tick')
        ).where(and_(
            table['turns'].c.branch == bindparam('branch'),
            table['turns'].c.turn == bindparam('turn')
        ))
    }
    for t in table.values():
        key = list(t.primary_key)
        if 'branch' in t.columns and 'turn' in t.columns and 'tick' in t.columns:
            branch = t.columns['branch']
            turn = t.columns['turn']
            tick = t.columns['tick']
            if branch in key and turn in key and tick in key:
                key = [branch, turn, tick]
                r[t.name + '_del_time'] = t.delete().where(and_(
                    t.c.branch == bindparam('branch'),
                    t.c.turn == bindparam('turn'),
                    t.c.tick == bindparam('tick')
                ))
        r[t.name + '_dump'] = select(list(t.c.values())).order_by(*key)
        r[t.name + '_insert'] = t.insert().values(tuple(bindparam(cname) for cname in t.c.keys()))
        r[t.name + '_count'] = select([func.COUNT()]).select_from(t)
        r[t.name + '_del'] = t.delete().where(and_(*[c == bindparam(c.name) for c in t.primary_key]))
    return r


def compile_sql(dialect, meta):
    from sqlalchemy.sql.ddl import CreateTable, CreateIndex
    r = {}
    table = tables_for_meta(meta)
    index = indices_for_table_dict(table)
    query = queries_for_table_dict(table)

    for t in table.values():
        r['create_' + t.name] = CreateTable(t).compile(dialect=dialect)
    for (tab, idx) in index.items():
        r['index_' + tab] = CreateIndex(idx).compile(dialect=dialect)
    for (name, q) in query.items():
        r[name] = q.compile(dialect=dialect)

    return r


class Alchemist(object):
    """Holds an engine and runs queries on it.

    """
    def __init__(self, engine):
        self.engine = engine
        self.conn = self.engine.connect()
        self.meta = MetaData()
        self.sql = compile_sql(self.engine.dialect, self.meta)
        def caller(k, *largs):
            statement = self.sql[k]
            if hasattr(statement, 'positiontup'):
                return self.conn.execute(statement, **dict(zip(statement.positiontup, largs)))
            elif largs:
                raise TypeError("{} is a DDL query, I think".format(k))
            return self.conn.execute(statement)
        def manycaller(k, *largs):
            statement = self.sql[k]
            return self.conn.execute(statement, *(dict(zip(statement.positiontup, larg)) for larg in largs))
        class Many(object):
            pass
        self.many = Many()
        for (key, query) in self.sql.items():
            setattr(self, key, partial(caller, key))
            setattr(self.many, key, partial(manycaller, key))


if __name__ == '__main__':
    from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite
    out = dict(
        (k, str(v)) for (k, v) in
        compile_sql(SQLiteDialect_pysqlite(), MetaData()).items()
    )

    print(dumps(out, indent=4, sort_keys=True))



# This file is part of allegedb, a database abstraction for versioned graphs
# Copyright (c) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Classes for in-memory storage and retrieval of historical graph data.
"""
from .window import WindowDict, HistoryError, FuturistWindowDict, TurnDict, SettingsTurnDict
from collections import OrderedDict, defaultdict, deque
from blinker import Signal


def _default_args_munger(self, k):
    """By default, :class:`PickyDefaultDict`'s ``type`` is instantiated with no positional arguments."""
    return tuple()


def _default_kwargs_munger(self, k):
    """By default, :class:`PickyDefaultDict`'s ``type`` is instantiated with no keyword arguments."""
    return {}


class PickyDefaultDict(dict):
    """A ``defaultdict`` alternative that requires values of a specific type.

    Pass some type object (such as a class) to the constructor to
    specify what type to use by default, which is the only type I will
    accept.

    Default values are constructed with no arguments by default;
    supply ``args_munger`` and/or ``kwargs_munger`` to override this.
    They take arguments ``self`` and the unused key being looked up.

    """
    __slots__ = ['type', 'args_munger', 'kwargs_munger', 'parent', 'key']

    def __init__(
            self, type=object,
            args_munger=_default_args_munger,
            kwargs_munger=_default_kwargs_munger
    ):
        self.type = type
        self.args_munger = args_munger
        self.kwargs_munger = kwargs_munger

    def __getitem__(self, k):
        if k in self:
            return super(PickyDefaultDict, self).__getitem__(k)
        try:
            ret = self[k] = self.type(
                *self.args_munger(self, k),
                **self.kwargs_munger(self, k)
            )
        except TypeError:
            raise KeyError
        return ret

    def _create(self, v):
        return self.type(v)

    def __setitem__(self, k, v):
        if type(v) is not self.type:
            v = self._create(v)
        super(PickyDefaultDict, self).__setitem__(k, v)


class StructuredDefaultDict(dict):
    """A ``defaultdict``-like class that expects values stored at a specific depth.

    Requires an integer to tell it how many layers deep to go.
    The innermost layer will be ``PickyDefaultDict``, which will take the
    ``type``, ``args_munger``, and ``kwargs_munger`` arguments supplied
    to my constructor.

    """
    __slots__ = ('layer', 'type', 'args_munger', 'kwargs_munger', 'parent', 'key', '_stuff')

    def __init__(
            self, layers, type=object,
            args_munger=_default_args_munger,
            kwargs_munger=_default_kwargs_munger
    ):
        if layers < 1:
            raise ValueError("Not enough layers")
        self.layer = layers
        self.type = type
        self.args_munger = args_munger
        self.kwargs_munger = kwargs_munger
        self._stuff = (layers, type, args_munger, kwargs_munger)

    def __getitem__(self, k):
        if k in self:
            return dict.__getitem__(self, k)
        layer, typ, args_munger, kwargs_munger = self._stuff
        if layer < 2:
            ret = PickyDefaultDict(
                typ, args_munger, kwargs_munger
            )
        else:
            ret = StructuredDefaultDict(
                layer-1, typ,
                args_munger, kwargs_munger
            )
        ret.parent = self
        ret.key = k
        dict.__setitem__(self, k, ret)
        return ret

    def __setitem__(self, k, v):
        if type(v) is StructuredDefaultDict:
            layer, typ, args_munger, kwargs_munger = self._stuff
            if (
                    v.layer == layer - 1
                    and v.type is typ
                    and v.args_munger is args_munger
                    and v.kwargs_munger is kwargs_munger
            ):
                super().__setitem__(k, v)
                return
        elif type(v) is PickyDefaultDict:
            layer, typ, args_munger, kwargs_munger = self._stuff
            if (
                layer < 2
                and v.type is typ
                and v.args_munger is args_munger
                and v.kwargs_munger is kwargs_munger
            ):
                super().__setitem__(k, v)
                return
        raise TypeError("Can't set layer {}".format(self.layer))


KEYCACHE_MAXSIZE = 1024


def lru_append(kc, lru, kckey, maxsize):
    """Delete old data from ``kc``, then add the new ``kckey``.

    :param kc: a three-layer keycache
    :param lru: an :class:`OrderedDict` with a key for each triple that should fill out ``kc``'s three layers
    :param kckey: a triple that indexes into ``kc``, which will be added to ``lru`` if needed
    :param maxsize: maximum number of entries in ``lru`` and, therefore, ``kc``

    """
    if kckey in lru:
        return
    while len(lru) >= maxsize:
        (peb, turn, tick), _ = lru.popitem(False)
        if peb not in kc:
            continue
        kcpeb = kc[peb]
        if turn not in kcpeb:
            continue
        kcpebturn = kcpeb[turn]
        if tick not in kcpebturn:
            continue
        del kcpebturn[tick]
        if not kcpebturn:
            del kcpeb[turn]
        if not kcpeb:
            del kc[peb]
    lru[kckey] = True


class Cache(Signal):
    """A data store that's useful for tracking graph revisions."""
    def __init__(self, db):
        super().__init__()
        self.db = db
        self.parents = StructuredDefaultDict(3, SettingsTurnDict)
        """Entity data keyed by the entities' parents.

        An entity's parent is what it's contained in. When speaking of a node,
        this is its graph. When speaking of an edge, the parent is usually the
        graph and the origin in a pair, though for multigraphs the destination
        might be part of the parent as well.

        Deeper layers of this cache are keyed by branch and revision.

        """
        self.keys = StructuredDefaultDict(2, SettingsTurnDict)
        """Cache of entity data keyed by the entities themselves.

        That means the whole tuple identifying the entity is the
        top-level key in this cache here. The second-to-top level
        is the key within the entity.

        Deeper layers of this cache are keyed by branch, turn, and tick.

        """
        self.keycache = PickyDefaultDict(SettingsTurnDict)
        """Keys an entity has at a given turn and tick."""
        self.branches = StructuredDefaultDict(1, SettingsTurnDict)
        """A less structured alternative to ``keys``.

        For when you already know the entity and the key within it,
        but still need to iterate through history to find the value.

        """
        self.shallowest = OrderedDict()
        """A dictionary for plain, unstructured hinting."""
        self.settings = PickyDefaultDict(SettingsTurnDict)
        """All the ``entity[key] = value`` operations that were performed on some turn"""
        self.presettings = PickyDefaultDict(SettingsTurnDict)
        """The values prior to ``entity[key] = value`` operations performed on some turn"""
        self.time_entity = {}
        self._kc_lru = OrderedDict()
        self._store_stuff = (
            self.parents, self.branches, self.keys, db.delete_plan,
            db._time_plan, self._iter_future_contradictions,
            db._branches, db._turn_end, self._store_journal,
            self.time_entity, db._where_cached, self.keycache
        )
        self._remove_stuff = (
            self.time_entity, self.parents, self.branches, self.keys,
            self.settings, self.presettings, self._remove_keycache, self.send
        )
        self._truncate_stuff = (
            self.parents, self.branches, self.keys, self.settings, self.presettings,
            self.keycache, self.send
        )

    def load(self, data):
        """Add a bunch of data. Must be in chronological order.

        But it doesn't need to all be from the same branch, as long as
        each branch is chronological of itself.

        """
        branches = defaultdict(list)
        for row in data:
            branches[row[-4]].append(row)
        # Make keycaches and valcaches. Must be done chronologically
        # to make forwarding work.
        childbranch = self.db._childbranch
        branch2do = deque(['trunk'])

        store = self._store
        while branch2do:
            branch = branch2do.popleft()
            for row in branches[branch]:
                store(*row, planning=False, loading=True)
            if branch in childbranch:
                branch2do.extend(childbranch[branch])

    def _valcache_lookup(self, cache, branch, turn, tick):
        """Return the value at the given time in ``cache``"""
        if branch in cache:
            branc = cache[branch]
            try:
                if turn in branc and branc[turn].rev_gettable(tick):
                    return branc[turn][tick]
                elif branc.rev_gettable(turn-1):
                    turnd = branc[turn-1]
                    return turnd[turnd.end]
            except HistoryError as ex:
                # probably shouldn't ever happen, empty branches shouldn't be kept in the cache at all...
                # but it's easy to handle
                if ex.deleted:
                    raise
        for b, r, t in self.db._iter_parent_btt(branch, turn, tick):
            if b in cache:
                if r in cache[b] and cache[b][r].rev_gettable(t):
                    try:
                        return cache[b][r][t]
                    except HistoryError as ex:
                        if ex.deleted:
                            raise
                elif cache[b].rev_gettable(r-1):
                    cbr = cache[b][r-1]
                    try:
                        return cbr[cbr.end]
                    except HistoryError as ex:
                        if ex.deleted:
                            raise

    def _get_keycachelike(self, keycache, keys, get_adds_dels, parentity, branch, turn, tick, *, forward):
        """Try to retrieve a frozenset representing extant keys.

        If I can't, generate one, store it, and return it.

        """
        keycache_key = parentity + (branch,)
        keycache2 = keycache3 = None
        if keycache_key in keycache:
            keycache2 = keycache[keycache_key]
            if turn in keycache2:
                keycache3 = keycache2[turn]
                if tick in keycache3:
                    return keycache3[tick]
        if forward:
            # Take valid values from the past of a keycache and copy them forward, into the present.
            # Assumes that time is only moving forward, never backward, never skipping any turns or ticks,
            # and any changes to the world state are happening through allegedb proper, meaning they'll all get cached.
            # In LiSE this means every change to the world state should happen inside of a call to
            # ``Engine.next_turn`` in a rule.
            if keycache2 and keycache2.rev_gettable(turn):
                # there's a keycache from a prior turn in this branch. Get it
                if turn not in keycache2:
                    # since it's not this *exact* turn there might be changes...
                    old_turn = keycache2.rev_before(turn)
                    old_turn_kc = keycache2[turn]
                    added, deleted = get_adds_dels(
                        keys[parentity], branch, turn, tick, stoptime=(
                            branch, old_turn, old_turn_kc.end
                        )
                    )
                    ret = old_turn_kc[old_turn_kc.end].union(added).difference(deleted)
                    # assert ret == get_adds_dels(keys[parentity], branch, turn, tick)[0]  # slow
                    new_turn_kc = WindowDict()
                    new_turn_kc[tick] = ret
                    keycache2[turn] = new_turn_kc
                    return ret
                if not keycache3:
                    keycache3 = keycache2[turn]
                if tick not in keycache3:
                    if keycache3.rev_gettable(tick):
                        added, deleted = get_adds_dels(
                            keys[parentity], branch, turn, tick, stoptime=(
                                branch, turn, keycache3.rev_before(tick)
                            )
                        )
                        ret = keycache3[tick].union(added).difference(deleted)
                        # assert ret == get_adds_dels(keys[parentity], branch, turn, tick)[0]  # slow
                        keycache3[tick] = ret
                        return ret
                    else:
                        turn_before = keycache2.rev_before(turn)
                        tick_before = keycache2[turn_before].end
                        keys_before = keycache2[turn_before][tick_before]
                        added, deleted = get_adds_dels(
                            keys[parentity], branch, turn, tick, stoptime=(
                                branch, turn_before, tick_before
                            )
                        )
                        ret = keycache3[tick] = keys_before.union(added).difference(deleted)
                        # assert ret == get_adds_dels(keys[parentity], branch, turn, tick)[0]  # slow
                        return ret
                # assert kcturn[tick] == get_adds_dels(keys[parentity], branch, turn, tick)[0]  # slow
                return keycache3[tick]
            else:
                for (parbranch, parturn, partick) in self.db._iter_parent_btt(branch, turn, tick):
                    par_kc_key = parentity + (parbranch,)
                    if par_kc_key in keycache:
                        kcpkc = keycache[par_kc_key]
                        if parturn in kcpkc and kcpkc[parturn].rev_gettable(partick):
                            parkeys = kcpkc[parturn][partick]
                            break
                        elif kcpkc.rev_gettable(parturn-1):
                            partkeys = kcpkc[parturn-1]
                            parkeys = partkeys[partkeys.end]
                            break
                else:
                    parkeys = frozenset()
                keycache2 = SettingsTurnDict()
                added, deleted = get_adds_dels(
                    keys[parentity], branch, turn, tick, stoptime=(
                        parbranch, parturn, partick
                    )
                )
                ret = parkeys.union(added).difference(deleted)
                keycache2[turn] = {tick: ret}
                keycache[keycache_key] = keycache2
                # assert ret == get_adds_dels(keys[parentity], branch, turn, tick)[0]  # slow
                return ret
        ret = frozenset(get_adds_dels(keys[parentity], branch, turn, tick)[0])
        if keycache2:
            if keycache3:
                keycache3[tick] = ret
            else:
                keycache2[turn] = {tick: ret}
        else:
            kcc = SettingsTurnDict()
            kcc[turn] = {tick: ret}
            keycache[keycache_key] = kcc
        return ret

    def _get_keycache(self, parentity, branch, turn, tick, *, forward):
        """Get a frozenset of keys that exist in the entity at the moment.

        With ``forward=True``, enable an optimization that copies old key sets
        forward and updates them.

        """
        lru_append(self.keycache, self._kc_lru, (parentity+(branch,), turn, tick), KEYCACHE_MAXSIZE)
        return self._get_keycachelike(
            self.keycache, self.keys, self._get_adds_dels,
            parentity, branch, turn, tick, forward=forward
        )

    def _update_keycache(self, *args, forward):
        """Add or remove a key in the set describing the keys that exist."""
        entity, key, branch, turn, tick, value = args[-6:]
        parent = args[:-6]
        kc = self._get_keycache(parent + (entity,), branch, turn, tick, forward=forward)
        if value is None:
            kc = kc.difference((key,))
        else:
            kc = kc.union((key,))
        self.keycache[parent+(entity, branch)][turn][tick] = kc

    def _get_adds_dels(self, cache, branch, turn, tick, *, stoptime=None):
        """Return a pair of ``(added, deleted)`` sets describing changes since ``stoptime``.

        With ``stoptime=None`` (the default), ``added`` will in fact be all keys.

        """
        added = set()
        deleted = set()
        for key, branches in cache.items():
            for (branc, trn, tck) in self.db._iter_parent_btt(branch, turn, tick, stoptime=stoptime):
                if branc not in branches or not branches[branc].rev_gettable(trn):
                    continue
                turnd = branches[branc]
                if trn in turnd:
                    if turnd[trn].rev_gettable(tck):
                        if turnd[trn][tck] is None:
                            deleted.add(key)
                        else:
                            added.add(key)
                        break
                    else:
                        trn -= 1
                if not turnd.rev_gettable(trn):
                    break
                tickd = turnd[trn]
                if tickd[tickd.end] is None:
                    deleted.add(key)
                else:
                    added.add(key)
                break
        return added, deleted

    def store(self, *args, planning=None, forward=None, loading=False, contra=True):
        """Put a value in various dictionaries for later .retrieve(...).

        Needs at least five arguments, of which the -1th is the value
        to store, the -2th is the tick to store it at, the -3th
        is the turn to store it in, the -4th is the branch the
        revision is in, the -5th is the key the value is for,
        and the remaining arguments identify the entity that has
        the key, eg. a graph, node, or edge.

        With ``planning=True``, you will be permitted to alter
        "history" that takes place after the last non-planning
        moment of time, without much regard to consistency.
        Otherwise, contradictions will be handled by deleting
        everything in the contradicted plan after the present moment,
        unless you set ``contra=False``.

        ``loading=True`` prevents me from updating the ORM's records
        of the ends of branches and turns.

        """
        db = self.db
        if planning is None:
            planning = db._planning
        if forward is None:
            forward = db._forward
        self._store(*args, planning=planning, loading=loading, contra=contra)
        if not db._no_kc:
            self._update_keycache(*args, forward=forward)
        self.send(self, key=args[-5], branch=args[-4], turn=args[-3], tick=args[-2], value=args[-1], action='store')

    def remove(self, branch, turn, tick):
        """Delete all data from a specific tick"""
        time_entity, parents, branches, keys, settings, presettings, remove_keycache, send = self._remove_stuff
        parent, entity, key = time_entity[branch, turn, tick]
        branchkey = parent + (entity, key)
        keykey = parent + (entity,)
        if parent in parents:
            parentt = parents[parent]
            if entity in parentt:
                entty = parentt[entity]
                if key in entty:
                    kee = entty[key]
                    if branch in kee:
                        branhc = kee[branch]
                        if turn in branhc:
                            trn = branhc[turn]
                            del trn[tick]
                            if not trn:
                                del branhc[turn]
                            if not branhc:
                                del kee[branch]
                    if not kee:
                        del entty[key]
                if not entty:
                    del parentt[entity]
            if not parentt:
                del parents[parent]
        if branchkey in branches:
            entty = branches[branchkey]
            if branch in entty:
                branhc = entty[branch]
                if turn in branhc:
                    trn = branhc[turn]
                    if tick in trn:
                        del trn[tick]
                    if not trn:
                        del branhc[turn]
                if not branhc:
                    del entty[branch]
            if not entty:
                del branches[branchkey]
        if keykey in keys:
            entty = keys[keykey]
            if key in entty:
                kee = entty[key]
                if branch in kee:
                    branhc = kee[branch]
                    if turn in branhc:
                        trn = entty[turn]
                        if tick in trn:
                            del trn[tick]
                        if not trn:
                            del branhc[turn]
                    if not branhc:
                        del kee[branch]
                if not kee:
                    del entty[key]
            if not entty:
                del keys[keykey]
        branhc = settings[branch]
        pbranhc = presettings[branch]
        trn = branhc[turn]
        ptrn = pbranhc[turn]
        if tick in trn:
            del trn[tick]
        if tick in ptrn:
            del ptrn[tick]
        if not ptrn:
            del pbranhc[turn]
            del branhc[turn]
        if not pbranhc:
            del settings[branch]
            del presettings[branch]
        self.shallowest = OrderedDict()
        remove_keycache(parent + (entity, branch), turn, tick)
        send(self, branch=branch, turn=turn, tick=tick, action='remove')

    def _remove_keycache(self, entity_branch, turn, tick):
        """Remove the future of a given entity from a branch in the keycache"""
        keycache = self.keycache
        if entity_branch in keycache:
            kc = keycache[entity_branch]
            if turn in kc:
                kcturn = kc[turn]
                if tick in kcturn:
                    del kcturn[tick]
                kcturn.truncate(tick)
                if not kcturn:
                    del kc[turn]
            kc.truncate(turn)
            if not kc:
                del keycache[entity_branch]

    def truncate(self, branch, turn, tick):
        """Delete all data after (not on) a specific tick"""
        parents, branches, keys, settings, presettings, keycache, send = self._truncate_stuff
        def truncate_branhc(branhc):
            if turn in branhc:
                trn = branhc[turn]
                trn.truncate(tick)
                branhc.truncate(turn)
                if not trn:
                    del branhc[turn]
            else:
                branhc.truncate(turn)
        for entities in parents.values():
            for keys in entities.values():
                for branches in keys.values():
                    if branch not in branches:
                        continue
                    truncate_branhc(branches[branch])
        for branches in branches.values():
            if branch not in branches:
                continue
            truncate_branhc(branches[branch])
        for keys in keys.values():
            for branches in keys.values():
                if branch not in branches:
                    continue
                truncate_branhc(branches[branch])
        truncate_branhc(settings[branch])
        truncate_branhc(presettings[branch])
        self.shallowest = OrderedDict()
        for entity_branch in keycache:
            if entity_branch[-1] == branch:
                truncate_branhc(keycache[entity_branch])
        send(self, branch=branch, turn=turn, tick=tick, action='truncate')

    @staticmethod
    def _iter_future_contradictions(entity, key, turns, branch, turn, tick, value):
        """If setting ``key=value`` would result in a contradiction, iterate over contradicted ``(turn, tick)``s."""
        # assumes that all future entries are in the plan
        if not turns:
            return
        if turn in turns:
            future_ticks = turns[turn].future(tick)
            for tck, newval in future_ticks.items():
                if newval != value:
                    yield turn, tck
            future_turns = turns.future(turn)
        elif turns.rev_gettable(turn):
            future_turns = turns.future(turn)
        else:
            future_turns = turns
        if not future_turns:
            return
        for trn, ticks in future_turns.items():
            for tick, newval in ticks.items():
                if newval != value:
                    yield trn, tick

    def _store(self, *args, planning, loading=False, contra=True):
        entity, key, branch, turn, tick, value = args[-6:]
        parent = args[:-6]
        (
            self_parents, self_branches, self_keys, delete_plan,
            time_plan, self_iter_future_contradictions,
            db_branches, db_turn_end, self_store_journal,
            self_time_entity, db_where_cached, keycache
        ) = self._store_stuff
        if parent:
            parentity = self_parents[parent][entity]
            if key in parentity:
                branches = parentity[key]
                turns = branches[branch]
            else:
                branches = self_branches[parent+(entity, key)] \
                    = self_keys[parent+(entity,)][key] \
                    = parentity[key]
                turns = branches[branch]
        else:
            if (entity, key) in self_branches:
                branches = self_branches[entity, key]
                turns = branches[branch]
            else:
                branches = self_branches[entity, key]
                self_keys[entity,][key] = branches
                turns = branches[branch]
        if planning:
            if turn in turns and tick < turns[turn].end:
                raise HistoryError(
                    "Already have some ticks after {} in turn {} of branch {}".format(
                        tick, turn, branch
                    )
                )
        if contra:
            contras = list(self_iter_future_contradictions(entity, key, turns, branch, turn, tick, value))
            if contras:
                self.shallowest = OrderedDict()
            for contra_turn, contra_tick in contras:
                if (branch, contra_turn, contra_tick) in time_plan:  # could've been deleted in this very loop
                    delete_plan(time_plan[branch, contra_turn, contra_tick])
        if not turns:
            branches[branch] = turns
        if not loading and not planning:
            parbranch, turn_start, tick_start, turn_end, tick_end = self.db._branches[branch]
            db_branches[branch] = parbranch, turn_start, tick_start, turn, tick
            db_turn_end[branch, turn] = tick
        self_store_journal(*args)
        self.shallowest[parent + (entity, key, branch, turn, tick)] = value
        shallowest = self.shallowest
        while len(shallowest) > KEYCACHE_MAXSIZE:
            shallowest.popitem(False)
        if turn in turns:
            the_turn = turns[turn]
            the_turn.truncate(tick)
            the_turn[tick] = value
        else:
            new = FuturistWindowDict()
            new[tick] = value
            turns[turn] = new
        self_time_entity[branch, turn, tick] = parent, entity, key
        where_cached = db_where_cached[args[-4:-1]]
        if self not in where_cached:
            where_cached.append(self)
        # if we're editing the past, have to invalidate the keycache
        keycache_key = parent + (entity, branch)
        if keycache_key in keycache:
            thiskeycache = keycache[keycache_key]
            if turn in thiskeycache:
                del thiskeycache[turn]
            thiskeycache.truncate(turn)
            if not thiskeycache:
                del keycache[keycache_key]

    def _store_journal(self, *args):
        # overridden in LiSE.cache.InitializedCache
        entity, key, branch, turn, tick, value = args[-6:]
        parent = args[:-6]
        settings_turns = self.settings[branch]
        presettings_turns = self.presettings[branch]
        prev = self._base_retrieve(args[:-1])
        if prev is KeyError:
            prev = None
        if turn in settings_turns or turn in settings_turns.future():
            # These assertions hold for most caches but not for the contents
            # caches, and are therefore commented out.
            # assert turn in presettings_turns or turn in presettings_turns.future()
            setticks = settings_turns[turn]
            # assert tick not in setticks
            presetticks = presettings_turns[turn]
            # assert tick not in presetticks
            presetticks[tick] = parent + (entity, key, prev)
            setticks[tick] = parent + (entity, key, value)
        else:
            presettings_turns[turn] = {tick: parent + (entity, key, prev)}
            settings_turns[turn] = {tick: parent + (entity, key, value)}

    def _base_retrieve(self, args):
        shallowest = self.shallowest
        if args in shallowest:
            return shallowest[args]
        entity = args[:-4]
        key, branch, turn, tick = args[-4:]
        branches = self.branches
        if entity+(key,) not in branches:
            return KeyError
        branchentk = branches[entity+(key,)]
        brancs = branchentk.get(branch)
        if brancs is not None and brancs.rev_gettable(turn):
            if turn in brancs:
                if brancs[turn].rev_gettable(tick):
                    ret = brancs[turn][tick]
                    shallowest[args] = ret
                    return ret
                elif brancs.rev_gettable(turn-1):
                    b1 = brancs[turn-1]
                    ret = b1[b1.end]
                    shallowest[args] = ret
                    return ret
            else:
                ret = brancs[turn]
                ret = ret[ret.end]
                shallowest[args] = ret
                return ret
        for (b, r, t) in self.db._iter_parent_btt(branch):
            brancs = branchentk.get(b)
            if brancs is not None and brancs.rev_gettable(r):
                if r in brancs and brancs[r].rev_gettable(t):
                    ret = brancs[r][t]
                elif brancs.rev_gettable(r-1):
                    ret = brancs[r-1]
                    ret = ret[ret.end]
                else:
                    continue
                shallowest[args] = ret
                return ret
        else:
            return KeyError

    def retrieve(self, *args):
        """Get a value previously .store(...)'d.

        Needs at least five arguments. The -1th is the tick
        within the turn you want,
        the -2th is that turn, the -3th is the branch,
        and the -4th is the key. All other arguments identify
        the entity that the key is in.

        """
        ret = self._base_retrieve(args)
        if ret is None:
            raise HistoryError("Set, then deleted", deleted=True)
        elif ret is KeyError:
            raise ret
        return ret

    def iter_entities_or_keys(self, *args, forward=None):
        """Iterate over the keys an entity has, if you specify an entity.

        Otherwise iterate over the entities themselves, or at any rate the
        tuple specifying which entity.

        """
        if forward is None:
            forward = self.db._forward
        entity = args[:-3]
        branch, turn, tick = args[-3:]
        if self.db._no_kc:
            yield from self._get_adds_dels(self.keys[entity], branch, turn, tick)[0]
            return
        yield from self._get_keycache(entity, branch, turn, tick, forward=forward)
    iter_entities = iter_keys = iter_entity_keys = iter_entities_or_keys

    def count_entities_or_keys(self, *args, forward=None):
        """Return the number of keys an entity has, if you specify an entity.

        Otherwise return the number of entities.

        """
        if forward is None:
            forward = self.db._forward
        entity = args[:-3]
        branch, turn, tick = args[-3:]
        if self.db._no_kc:
            return len(self._get_adds_dels(self.keys[entity], branch, turn, tick)[0])
        return len(self._get_keycache(entity, branch, turn, tick, forward=forward))
    count_entities = count_keys = count_entity_keys = count_entities_or_keys

    def contains_entity_or_key(self, *args):
        """Check if an entity has a key at the given time, if entity specified.

        Otherwise check if the entity exists.

        """
        try:
            return self.retrieve(*args) is not None
        except KeyError:
            return False
    contains_entity = contains_key = contains_entity_key \
                    = contains_entity_or_key


class NodesCache(Cache):
    """A cache for remembering whether nodes exist at a given time."""

    def _store(self, graph, node, branch, turn, tick, ex, *, planning, loading=False, contra=True):
        if not ex:
            ex = None
        return super()._store(graph, node, branch, turn, tick, ex, planning=planning, loading=loading, contra=contra)

    def _update_keycache(self, *args, forward):
        graph, node, branch, turn, tick, ex = args
        if not ex:
            ex = None
        super()._update_keycache(graph, node, branch, turn, tick, ex, forward=forward)

    def _iter_future_contradictions(self, entity, key, turns, branch, turn, tick, value):
        yield from super()._iter_future_contradictions(entity, key, turns, branch, turn, tick, value)
        yield from self.db._edges_cache._slow_iter_node_contradicted_times(
            branch, turn, tick, entity, key
        )


class EdgesCache(Cache):
    """A cache for remembering whether edges exist at a given time."""
    @property
    def successors(self):
        return self.parents

    def __init__(self, db):
        Cache.__init__(self, db)
        self.destcache = PickyDefaultDict(SettingsTurnDict)
        self.origcache = PickyDefaultDict(SettingsTurnDict)
        self.predecessors = StructuredDefaultDict(3, TurnDict)
        self._origcache_lru = OrderedDict()
        self._destcache_lru = OrderedDict()
        self._get_destcache_stuff = (
            self.destcache, self._destcache_lru, self._get_keycachelike,
            self.successors, self._adds_dels_sucpred
        )
        self._get_origcache_stuff = (
            self.origcache, self._origcache_lru, self._get_keycachelike,
            self.predecessors, self._adds_dels_sucpred
        )

    def _slow_iter_node_contradicted_times(self, branch, turn, tick, graph, node):
        # slow and bad.
        retrieve = self._base_retrieve
        for items in (self.successors[graph, node].items(), self.predecessors[graph, node].items()):
            for dest, idxs in items:  # dest might really be orig
                for idx, branches in idxs.items():
                    brnch = branches[branch]
                    if turn in brnch:
                        ticks = brnch[turn]
                        for tck, present in ticks.future(tick).items():
                            if tck > tick and present is not retrieve((graph, node, dest, idx, branch, turn, tick)):
                                yield turn, tck
                    for trn, ticks in brnch.future(turn).items():
                        for tck, present in ticks.items():
                            if present is not retrieve((graph, node, dest, idx, branch, turn, tick)):
                                yield trn, tck

    def _adds_dels_sucpred(self, cache, branch, turn, tick, *, stoptime=None):
        """Take the successors or predecessors cache and get nodes added or deleted from it

        Operates like ``_get_adds_dels``.

        """
        added = set()
        deleted = set()
        for node, nodes in cache.items():
            addidx, delidx = self._get_adds_dels(nodes, branch, turn, tick, stoptime=stoptime)
            if addidx and not delidx:
                added.add(node)
            elif delidx and not addidx:
                deleted.add(node)
        return added, deleted

    def _get_destcache(self, graph, orig, branch, turn, tick, *, forward):
        """Return a set of destination nodes succeeding ``orig``"""
        destcache, destcache_lru, get_keycachelike, successors, adds_dels_sucpred = self._get_destcache_stuff
        lru_append(destcache, destcache_lru, ((graph, orig, branch), turn, tick), KEYCACHE_MAXSIZE)
        return get_keycachelike(
            destcache, successors, adds_dels_sucpred, (graph, orig),
            branch, turn, tick, forward=forward
        )

    def _get_origcache(self, graph, dest, branch, turn, tick, *, forward):
        """Return a set of origin nodes leading to ``dest``"""
        origcache, origcache_lru, get_keycachelike, predecessors, adds_dels_sucpred = self._get_origcache_stuff
        lru_append(origcache, origcache_lru, ((graph, dest, branch), turn, tick), KEYCACHE_MAXSIZE)
        return get_keycachelike(
            origcache, predecessors, adds_dels_sucpred, (graph, dest),
            branch, turn, tick, forward=forward
        )

    def iter_successors(self, graph, orig, branch, turn, tick, *, forward=None):
        """Iterate over successors of a given origin node at a given time."""
        if self.db._no_kc:
            yield from self._adds_dels_sucpred(self.successors[graph, orig], branch, turn, tick)[0]
            return
        if forward is None:
            forward = self.db._forward
        yield from self._get_destcache(graph, orig, branch, turn, tick, forward=forward)

    def iter_predecessors(self, graph, dest, branch, turn, tick, *, forward=None):
        """Iterate over predecessors to a given destination node at a given time."""
        if self.db._no_kc:
            yield from self._adds_dels_sucpred(self.predecessors[graph, dest], branch, turn, tick)[0]
            return
        if forward is None:
            forward = self.db._forward
        yield from self._get_origcache(graph, dest, branch, turn, tick, forward=forward)

    def count_successors(self, graph, orig, branch, turn, tick, *, forward=None):
        """Return the number of successors to a given origin node at a given time."""
        if self.db._no_kc:
            return len(self._adds_dels_sucpred(self.successors[graph, orig], branch, turn, tick)[0])
        if forward is None:
            forward = self.db._forward
        return len(self._get_destcache(graph, orig, branch, turn, tick, forward=forward))

    def count_predecessors(self, graph, dest, branch, turn, tick, *, forward=None):
        """Return the number of predecessors from a given destination node at a given time."""
        if self.db._no_kc:
            return len(self._adds_dels_sucpred(self.predecessors[graph, dest], branch, turn, tick)[0])
        if forward is None:
            forward = self.db._forward
        return len(self._get_origcache(graph, dest, branch, turn, tick, forward=forward))

    def has_successor(self, graph, orig, dest, branch, turn, tick, *, forward=None):
        """Return whether an edge connects the origin to the destination at the given time.

        Doesn't require the edge's index, which makes it slower than retrieving a
        particular edge.

        """
        if forward is None:
            forward = self.db._forward
        return dest in self._get_destcache(graph, orig, branch, turn, tick, forward=forward)

    def has_predecessor(self, graph, dest, orig, branch, turn, tick, *, forward=None):
        """Return whether an edge connects the destination to the origin at the given time.

        Doesn't require the edge's index, which makes it slower than retrieving a
        particular edge.

        """
        if forward is None:
            forward = self.db._forward
        return orig in self._get_origcache(graph, dest, branch, turn, tick, forward=forward)

    def _store(self, graph, orig, dest, idx, branch, turn, tick, ex, *, planning=None, loading=False, contra=True):
        if not ex:
            ex = None
        if planning is None:
            planning = self.db.planning
        Cache._store(self, graph, orig, dest, idx, branch, turn, tick, ex, planning=planning, loading=loading, contra=contra)
        self.predecessors[(graph, dest)][orig][idx][branch][turn] \
            = self.successors[graph, orig][dest][idx][branch][turn]
        # if ex:
        #     assert self.retrieve(graph, orig, dest, idx, branch, turn, tick)
        #     assert self.has_successor(graph, orig, dest, branch, turn, tick)
        #     assert self.has_predecessor(graph, dest, orig, branch, turn, tick)
        # else:
        #     assert self._base_retrieve((graph, orig, dest, idx, branch, turn, tick)) in (None, KeyError)
        #     assert not self.has_successor(graph, orig, dest, branch, turn, tick)
        #     assert not self.has_predecessor(graph, dest, orig, branch, turn, tick)


# This file is part of allegedb, an object relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The main interface to the allegedb ORM, and some supporting functions and classes"""
from contextlib import ContextDecorator
from weakref import WeakValueDictionary

from blinker import Signal

from allegedb.window import update_window, update_backward_window
from .graph import (
    Graph,
    DiGraph,
    MultiGraph,
    MultiDiGraph,
    Node,
    Edge,
    GraphsMapping
)
from .query import QueryEngine, TimeError
from .window import HistoryError


class GraphNameError(KeyError):
    """For errors involving graphs' names"""


class PlanningContext(ContextDecorator):
    """A context manager for 'hypothetical' edits.

    Start a block of code like:

    ```
    with orm.plan():
        ...
    ```

    and any changes you make to the world state within that block will be
    'plans,' meaning that they are used as defaults. The world will
    obey your plan unless you make changes to the same entities outside
    of the plan, in which case the world will obey those, and cancel any
    future plan.

    New branches cannot be started within plans. The ``with orm.forward():``
    optimization is disabled within a ``with orm.plan():`` block, so
    consider another approach instead of making a very large plan.

    """
    __slots__ = ['orm', 'id', 'forward']

    def __init__(self, orm):
        self.orm = orm

    def __enter__(self):
        orm = self.orm
        if orm._planning:
            raise ValueError("Already planning")
        orm._planning = True
        branch, turn, tick = orm._btt()
        self.id = myid = orm._last_plan = orm._last_plan + 1
        self.forward = orm._forward
        if orm._forward:
            orm._forward = False
        orm._plans[myid] = branch, turn, tick
        orm._plans_uncommitted.append((myid, branch, turn, tick))
        orm._branches_plans[branch].add(myid)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.orm._planning = False
        if self.forward:
            self.orm._forward = True


class TimeSignal:
    """Acts like a tuple of ``(branch, turn)`` for the most part.

    This wraps a ``Signal``. To set a function to be called whenever the branch
    or turn changes, pass it to my ``connect`` method.

    """
    def __init__(self, engine, sig):
        self.engine = engine
        self.branch = self.engine.branch
        self.turn = self.engine.turn
        self.sig = sig

    def __iter__(self):
        yield self.branch
        yield self.turn

    def __len__(self):
        return 2

    def __getitem__(self, i):
        if i in ('branch', 0):
            return self.branch
        if i in ('turn', 1):
            return self.turn
        raise IndexError

    def connect(self, *args, **kwargs):
        self.sig.connect(*args, **kwargs)

    def send(self, *args, **kwargs):
        self.sig.send(*args, **kwargs)

    def __str__(self):
        return str(tuple(self))

    def __eq__(self, other):
        return tuple(self) == other

    def __ne__(self, other):
        return tuple(self) != other

    def __gt__(self, other):
        return tuple(self) > other

    def __ge__(self, other):
        return tuple(self) >= other

    def __lt__(self, other):
        return tuple(self) < other

    def __le__(self, other):
        return tuple(self) <= other


class TimeSignalDescriptor:
    __doc__ = TimeSignal.__doc__
    signals = {}

    def __get__(self, inst, cls):
        if id(inst) not in self.signals:
            self.signals[id(inst)] = Signal()
        return TimeSignal(inst, self.signals[id(inst)])

    def __set__(self, inst, val):
        if id(inst) not in self.signals:
            self.signals[id(inst)] = Signal()
        sig = self.signals[id(inst)]
        branch_then, turn_then, tick_then = inst._btt()
        branch_now, turn_now = val
        if (branch_then, turn_then) == (branch_now, turn_now):
            return
        e = inst
        # enforce the arrow of time, if it's in effect
        if e._forward and not e._planning:
            if branch_now != branch_then:
                raise TimeError("Can't change branches in a forward context")
            if turn_now < turn_then:
                raise TimeError("Can't time travel backward in a forward context")
            if turn_now > turn_then + 1:
                raise TimeError("Can't skip turns in a forward context")
        # make sure I'll end up within the revision range of the
        # destination branch
        branches = e._branches

        if branch_now in branches:
            tick_now = e._turn_end_plan.setdefault(
                (branch_now, turn_now),
                tick_then
            )
            parent, turn_start, tick_start, turn_end, tick_end = branches[branch_now]
            if turn_now < turn_start:
                raise ValueError(
                    "The turn number {} "
                    "occurs before the start of "
                    "the branch {}".format(turn_now, branch_now)
                )
            if turn_now == turn_start and tick_now < tick_start:
                raise ValueError(
                    "The tick number {}"
                    "on turn {} "
                    "occurs before the start of "
                    "the branch {}".format(
                        tick_now, turn_now, branch_now
                    )
                )
            if not e._planning and (
                turn_now > turn_end or (
                    turn_now == turn_end and tick_now > tick_end
                )
            ):
                branches[branch_now] = parent, turn_start, tick_start, turn_now, tick_now
        else:
            tick_now = tick_then
            branches[branch_now] = (
                branch_then, turn_now, tick_now, turn_now, tick_now
            )
            e.query.new_branch(branch_now, branch_then, turn_now, tick_now)
        e._obranch, e._oturn = val

        if not e._planning:
            if tick_now > e._turn_end[val]:
                e._turn_end[val] = tick_now
        e._otick = e._turn_end_plan[val] = tick_now
        sig.send(
            e,
            branch_then=branch_then,
            turn_then=turn_then,
            tick_then=tick_then,
            branch_now=branch_now,
            turn_now=turn_now,
            tick_now=tick_now
        )


def setgraphval(delta, graph, key, val):
    """Change a delta to say that a graph stat was set to a certain value"""
    delta.setdefault(graph, {})[key] = val


def setnode(delta, graph, node, exists):
    """Change a delta to say that a node was created or deleted"""
    delta.setdefault(graph, {}).setdefault('nodes', {})[node] = bool(exists)


def setnodeval(delta, graph, node, key, value):
    """Change a delta to say that a node stat was set to a certain value"""
    if (
        graph in delta and 'nodes' in delta[graph] and
        node in delta[graph]['nodes'] and not delta[graph]['nodes'][node]
    ):
        return
    delta.setdefault(graph, {}).setdefault('node_val', {}).setdefault(node, {})[key] = value


def setedge(delta, is_multigraph, graph, orig, dest, idx, exists):
    """Change a delta to say that an edge was created or deleted"""
    if is_multigraph(graph):
        delta.setdefault(graph, {}).setdefault('edges', {})\
            .setdefault(orig, {}).setdefault(dest, {})[idx] = bool(exists)
    else:
        delta.setdefault(graph, {}).setdefault('edges', {})\
            .setdefault(orig, {})[dest] = bool(exists)


def setedgeval(delta, is_multigraph, graph, orig, dest, idx, key, value):
    """Change a delta to say that an edge stat was set to a certain value"""
    if is_multigraph(graph):
        if (
            graph in delta and 'edges' in delta[graph] and
            orig in delta[graph]['edges'] and dest in delta[graph]['edges'][orig]
            and idx in delta[graph]['edges'][orig][dest]
            and not delta[graph]['edges'][orig][dest][idx]
        ):
            return
        delta.setdefault(graph, {}).setdefault('edge_val', {})\
            .setdefault(orig, {}).setdefault(dest, {})\
            .setdefault(idx, {})[key] = value
    else:
        if (
                                    graph in delta and 'edges' in delta[graph] and
                                orig in delta[graph]['edges'] and dest in delta[graph]['edges'][orig]
                and not delta[graph]['edges'][orig][dest]
        ):
            return
        delta.setdefault(graph, {}).setdefault('edge_val', {})\
            .setdefault(orig, {}).setdefault(dest, {})[key] = value


class ORM(object):
    """Instantiate this with the same string argument you'd use for a
    SQLAlchemy ``create_engine`` call. This will be your interface to
    allegedb.

    """
    node_cls = Node
    edge_cls = Edge
    query_engine_cls = QueryEngine
    illegal_graph_names = ['global']
    illegal_node_names = ['nodes', 'node_val', 'edges', 'edge_val']
    time = TimeSignalDescriptor()

    def _make_node(self, graph, node):
        return self.node_cls(graph, node)

    def _get_node(self, graph, node):
        node_objs, node_exists, make_node = self._get_node_stuff
        if type(graph) is str:
            graphn = graph
            graph = self.graph[graphn]
        else:
            graphn = graph.name
        key = (graphn, node)
        if key in node_objs:
            return node_objs[key]
        if not node_exists(graphn, node):
            raise KeyError("No such node: {} in {}".format(node, graphn))
        ret = make_node(graph, node)
        node_objs[key] = ret
        return ret

    def _make_edge(self, graph, orig, dest, idx):
        return self.edge_cls(graph, orig, dest, idx)

    def _get_edge(self, graph, orig, dest, idx=0):
        edge_objs, edge_exists, make_edge = self._get_edge_stuff
        if type(graph) is str:
            graphn = graph
            graph = self.graph[graphn]
        else:
            graphn = graph.name
        key = (graphn, orig, dest, idx)
        if key in edge_objs:
            return edge_objs[key]
        if not edge_exists(graphn, orig, dest, idx):
            raise KeyError("No such edge: {}->{}[{}] in {}".format(orig, dest, idx, graphn))
        ret = make_edge(graph, orig, dest, idx)
        edge_objs[key] = ret
        return ret

    def plan(self):
        return PlanningContext(self)
    plan.__doc__ = PlanningContext.__doc__

    from contextlib import contextmanager

    @contextmanager
    def advancing(self):
        """A context manager for when time is moving forward one turn at a time.

        When used in LiSE, this means that the game is being simulated.
        It changes how the caching works, making it more efficient.

        """
        if self._forward:
            raise ValueError("Already advancing")
        self._forward = True
        yield
        self._forward = False

    @contextmanager
    def batch(self):
        """A context manager for when you're creating lots of state.

        Reads will be much slower in a batch, but writes will be faster.

        You *can* combine this with ``advancing`` but it isn't any faster.

        """
        if self._no_kc:
            raise ValueError("Already in a batch")
        self._no_kc = True
        yield
        self._no_kc = False

    def get_delta(self, branch, turn_from, tick_from, turn_to, tick_to):
        """Get a dictionary describing changes to all graphs.

        The keys are graph names. Their values are dictionaries of the graphs'
        attributes' new values, with ``None`` for deleted keys. Also in those graph
        dictionaries are special keys 'node_val' and 'edge_val' describing changes
        to node and edge attributes, and 'nodes' and 'edges' full of booleans
        indicating whether a node or edge exists.

        """
        from functools import partial
        if turn_from == turn_to:
            return self.get_turn_delta(branch, turn_from, tick_from, tick_to)
        delta = {}
        graph_objs = self._graph_objs
        if turn_to < turn_from:
            updater = partial(update_backward_window, turn_from, tick_from, turn_to, tick_to)
            gvbranches = self._graph_val_cache.presettings
            nbranches = self._nodes_cache.presettings
            nvbranches = self._node_val_cache.presettings
            ebranches = self._edges_cache.presettings
            evbranches = self._edge_val_cache.presettings
        else:
            updater = partial(update_window, turn_from, tick_from, turn_to, tick_to)
            gvbranches = self._graph_val_cache.settings
            nbranches = self._nodes_cache.settings
            nvbranches = self._node_val_cache.settings
            ebranches = self._edges_cache.settings
            evbranches = self._edge_val_cache.settings

        if branch in gvbranches:
            updater(partial(setgraphval, delta), gvbranches[branch])

        if branch in nbranches:
            updater(partial(setnode, delta), nbranches[branch])

        if branch in nvbranches:
            updater(partial(setnodeval, delta), nvbranches[branch])

        if branch in ebranches:
            updater(partial(setedge, delta, lambda g: graph_objs[g].is_multigraph()), ebranches[branch])

        if branch in evbranches:
            updater(partial(setedgeval, delta, lambda g: graph_objs[g].is_multigraph()), evbranches[branch])

        return delta

    def get_turn_delta(self, branch=None, turn=None, tick_from=0, tick_to=None):
        """Get a dictionary describing changes made on a given turn.

        If ``tick_to`` is not supplied, report all changes after ``tick_from``
        (default 0).

        The keys are graph names. Their values are dictionaries of the graphs'
        attributes' new values, with ``None`` for deleted keys. Also in those graph
        dictionaries are special keys 'node_val' and 'edge_val' describing changes
        to node and edge attributes, and 'nodes' and 'edges' full of booleans
        indicating whether a node or edge exists.

        :arg branch: A branch of history; defaults to the current branch
        :arg turn: The turn in the branch; defaults to the current turn
        :arg tick_from: Starting tick; defaults to 0

        """
        branch = branch or self.branch
        turn = turn or self.turn
        tick_to = tick_to or self.tick
        delta = {}
        if tick_from < tick_to:
            gvbranches = self._graph_val_cache.settings
            nbranches = self._nodes_cache.settings
            nvbranches = self._node_val_cache.settings
            ebranches = self._edges_cache.settings
            evbranches = self._edge_val_cache.settings
        else:
            gvbranches = self._graph_val_cache.presettings
            nbranches = self._nodes_cache.presettings
            nvbranches = self._node_val_cache.presettings
            ebranches = self._edges_cache.presettings
            evbranches = self._edge_val_cache.presettings

        if branch in gvbranches and turn in gvbranches[branch]:
            for graph, key, value in gvbranches[branch][turn][tick_from:tick_to]:
                if graph in delta:
                    delta[graph][key] = value
                else:
                    delta[graph] = {key: value}

        if branch in nbranches and turn in nbranches[branch]:
            for graph, node, exists in nbranches[branch][turn][tick_from:tick_to]:
                delta.setdefault(graph, {}).setdefault('nodes', {})[node] = bool(exists)

        if branch in nvbranches and turn in nvbranches[branch]:
            for graph, node, key, value in nvbranches[branch][turn][tick_from:tick_to]:
                if (
                    graph in delta and 'nodes' in delta[graph] and
                    node in delta[graph]['nodes'] and not delta[graph]['nodes'][node]
                ):
                    continue
                nodevd = delta.setdefault(graph, {}).setdefault('node_val', {})
                if node in nodevd:
                    nodevd[node][key] = value
                else:
                    nodevd[node] = {key: value}

        graph_objs = self._graph_objs
        if branch in ebranches and turn in ebranches[branch]:
            for graph, orig, dest, idx, exists in ebranches[branch][turn][tick_from:tick_to]:
                if graph_objs[graph].is_multigraph():
                    if (
                        graph in delta and 'edges' in delta[graph] and
                        orig in delta[graph]['edges'] and dest in delta[graph]['edges'][orig]
                        and idx in delta[graph]['edges'][orig][dest]
                        and not delta[graph]['edges'][orig][dest][idx]
                    ):
                        continue
                    delta.setdefault(graph, {}).setdefault('edges', {})\
                        .setdefault(orig, {}).setdefault(dest, {})[idx] = bool(exists)
                else:
                    if (
                        graph in delta and 'edges' in delta[graph] and
                        orig in delta[graph]['edges'] and dest in delta[graph]['edges'][orig]
                        and not delta[graph]['edges'][orig][dest]
                    ):
                        continue
                    delta.setdefault(graph, {}).setdefault('edges', {})\
                        .setdefault(orig, {})[dest] = bool(exists)

        if branch in evbranches and turn in evbranches[branch]:
            for graph, orig, dest, idx, key, value in evbranches[branch][turn][tick_from:tick_to]:
                edgevd = delta.setdefault(graph, {}).setdefault('edge_val', {})\
                    .setdefault(orig, {}).setdefault(dest, {})
                if graph_objs[graph].is_multigraph():
                    if idx in edgevd:
                        edgevd[idx][key] = value
                    else:
                        edgevd[idx] = {key: value}
                else:
                    edgevd[key] = value

        return delta

    def _init_caches(self):
        from collections import defaultdict
        from .cache import Cache, NodesCache, EdgesCache
        self._where_cached = defaultdict(list)
        self._global_cache = self.query._global_cache = {}
        self._node_objs = node_objs = WeakValueDictionary()
        self._get_node_stuff = (node_objs, self._node_exists, self._make_node)
        self._edge_objs = edge_objs = WeakValueDictionary()
        self._get_edge_stuff = (edge_objs, self._edge_exists, self._make_edge)
        for k, v in self.query.global_items():
            if k == 'branch':
                self._obranch = v
            elif k == 'turn':
                self._oturn = int(v)
            elif k == 'tick':
                self._otick = int(v)
            else:
                self._global_cache[k] = v
        self._childbranch = defaultdict(set)
        """Immediate children of a branch"""
        self._branches = {}
        """Start time, end time, and parent of each branch"""
        self._branch_parents = defaultdict(set)
        """Parents of a branch at any remove"""
        self._turn_end = defaultdict(lambda: 0)
        """Tick on which a (branch, turn) ends"""
        self._turn_end_plan = defaultdict(lambda: 0)
        """Tick on which a (branch, turn) ends, even if it hasn't been simulated"""
        self._graph_objs = {}
        self._plans = {}
        self._branches_plans = defaultdict(set)
        self._plan_ticks = defaultdict(lambda: defaultdict(list))
        self._time_plan = {}
        self._plans_uncommitted = []
        self._plan_ticks_uncommitted = []
        self._graph_val_cache = Cache(self)
        self._graph_val_cache.setdb = self.query.graph_val_set
        self._graph_val_cache.deldb = self.query.graph_val_del_time
        self._nodes_cache = NodesCache(self)
        self._nodes_cache.setdb = self.query.exist_node
        self._nodes_cache.deldb = self.query.nodes_del_time
        self._edges_cache = EdgesCache(self)
        self._edges_cache.setdb = self.query.exist_edge
        self._edges_cache.deldb = self.query.edges_del_time
        self._node_val_cache = Cache(self)
        self._node_val_cache.setdb = self.query.node_val_set
        self._node_val_cache.deldb = self.query.node_val_del_time
        self._edge_val_cache = Cache(self)
        self._edge_val_cache.setdb = self.query.edge_val_set
        self._edge_val_cache.deldb = self.query.edge_val_del_time

    def _load_graphs(self):
        for (graph, typ) in self.query.graphs_types():
            self._graph_objs[graph] = {
                'Graph': Graph,
                'DiGraph': DiGraph,
                'MultiGraph': MultiGraph,
                'MultiDiGraph': MultiDiGraph
            }[typ](self, graph)

    def __init__(
            self,
            dbstring,
            alchemy=True,
            connect_args={},
            validate=False
    ):
        """Make a SQLAlchemy engine if possible, else a sqlite3 connection. In
        either case, begin a transaction.

        :arg dbstring: rfc1738 URL for a database connection. Unless it begins with
        "sqlite:///", SQLAlchemy will be required.
        :arg alchemy: Set to ``False`` to use the precompiled SQLite queries even if
        SQLAlchemy is available.
        :arg connect_args: Dictionary of keyword arguments to be used for the database
        connection.
        :arg validate: Whether to perform an integrity test on the data.

        """
        self._planning = False
        self._forward = False
        self._no_kc = False
        if not hasattr(self, 'query'):
            self.query = self.query_engine_cls(
                dbstring, connect_args, alchemy,
                getattr(self, 'pack', None), getattr(self, 'unpack', None)
            )
        self.query.initdb()
        # in case this is the first startup
        self._obranch = 'trunk'
        self._otick = self._oturn = 0
        self._init_caches()
        for (branch, parent, parent_turn, parent_tick, end_turn, end_tick) in self.query.all_branches():
            self._branches[branch] = (parent, parent_turn, parent_tick, end_turn, end_tick)
            self._upd_branch_parentage(parent, branch)
        for (branch, turn, end_tick, plan_end_tick) in self.query.turns_dump():
            self._turn_end[branch, turn] = end_tick
            self._turn_end_plan[branch, turn] = plan_end_tick
        if 'trunk' not in self._branches:
            self._branches['trunk'] = None, 0, 0, 0, 0
        self._load_graphs()
        self._init_load(validate=validate)

    def _upd_branch_parentage(self, parent, child):
        self._childbranch[parent].add(child)
        self._branch_parents[child].add(parent)
        while parent in self._branches:
            parent, _, _, _, _ = self._branches[parent]
            self._branch_parents[child].add(parent)

    def _init_load(self, validate=False):
        if not hasattr(self, 'graph'):
            self.graph = GraphsMapping(self)
        noderows = [
            (graph, node, branch, turn, tick, ex if ex else None)
            for (graph, node, branch, turn, tick, ex)
            in self.query.nodes_dump()
        ]
        self._nodes_cache.load(noderows)
        edgerows = [
            (graph, orig, dest, idx, branch, turn, tick, ex if ex else None)
            for (graph, orig, dest, idx, branch, turn, tick, ex)
            in self.query.edges_dump()
        ]
        self._edges_cache.load(edgerows)
        self._graph_val_cache.load(self.query.graph_val_dump())
        self._node_val_cache.load(self.query.node_val_dump())
        self._edge_val_cache.load(self.query.edge_val_dump())
        last_plan = -1
        plans = self._plans
        branches_plans = self._branches_plans
        for plan, branch, turn, tick in self.query.plans_dump():
            plans[plan] = branch, turn, tick
            branches_plans[branch].add(plan)
            if plan > last_plan:
                last_plan = plan
        self._last_plan = last_plan
        plan_ticks = self._plan_ticks
        time_plan = self._time_plan
        for plan, turn, tick in self.query.plan_ticks_dump():
            plan_ticks[plan][turn].append(tick)
            time_plan[plans[plan][0], turn, tick] = plan

    def __enter__(self):
        """Enable the use of the ``with`` keyword"""
        return self

    def __exit__(self, *args):
        """Alias for ``close``"""
        self.close()

    def is_parent_of(self, parent, child):
        """Return whether ``child`` is a branch descended from ``parent`` at
        any remove.

        """
        if parent == 'trunk':
            return True
        if child == 'trunk':
            return False
        if child not in self._branches:
            raise ValueError(
                "The branch {} seems not to have ever been created".format(
                    child
                )
            )
        if self._branches[child][0] == parent:
            return True
        return self.is_parent_of(parent, self._branches[child][0])

    def _get_branch(self):
        return self._obranch

    def _set_branch(self, v):
        if self._planning:
            raise ValueError("Don't change branches while planning")
        curbranch, curturn, curtick = self._btt()
        if curbranch == v:
            self._otick = self._turn_end_plan[curbranch, curturn]
            return
        # make sure I'll end up within the revision range of the
        # destination branch
        if v != 'trunk' and v in self._branches:
            parturn = self._branches[v][1]
            if curturn < parturn:
                raise ValueError(
                    "Tried to jump to branch {br}, "
                    "which starts at turn {rv}. "
                    "Go to turn {rv} or later to use this branch.".format(
                        br=v,
                        rv=parturn
                    )
                )
        branch_is_new = v not in self._branches
        if branch_is_new:
            # assumes the present turn in the parent branch has
            # been finalized.
            self.query.new_branch(v, curbranch, curturn, curtick)
            self._branches[v] = curbranch, curturn, curtick, curturn, curtick
            self._upd_branch_parentage(v, curbranch)
            self._turn_end_plan[v, curturn] = self._turn_end[v, curturn] = curtick
        self._obranch = v
        self._otick = self._turn_end_plan[v, curturn]
        if branch_is_new:
            self._copy_plans(curbranch, curturn, curtick)

    def _copy_plans(self, branch_from, turn_from, tick_from):
        """Collect all plans that are active at the given time and copy them to the current branch"""
        plan_ticks = self._plan_ticks
        plan_ticks_uncommitted = self._plan_ticks_uncommitted
        time_plan = self._time_plan
        plans = self._plans
        branch = self.branch
        where_cached = self._where_cached
        last_plan = self._last_plan
        turn_end_plan = self._turn_end_plan
        for plan_id in self._branches_plans[branch_from]:
            _, start_turn, start_tick = plans[plan_id]
            if start_turn > turn_from or (start_turn == turn_from and start_tick > tick_from):
                continue
            incremented = False
            for turn, ticks in list(plan_ticks[plan_id].items()):
                if turn < turn_from:
                    continue
                for tick in ticks:
                    if turn == turn_from and tick < tick_from:
                        continue
                    if not incremented:
                        self._last_plan = last_plan = last_plan + 1
                        incremented = True
                        plans[last_plan] = branch, turn, tick
                    for cache in where_cached[branch_from, turn, tick]:
                        data = cache.settings[branch_from][turn][tick]
                        value = data[-1]
                        key = data[:-1]
                        args = key + (branch, turn, tick, value)
                        if hasattr(cache, 'setdb'):
                            cache.setdb(*args)
                        cache.store(*args, planning=True)
                        plan_ticks[last_plan][turn].append(tick)
                        plan_ticks_uncommitted.append((last_plan, turn, tick))
                        time_plan[branch, turn, tick] = last_plan
                        turn_end_plan[branch, turn] = tick

    def delete_plan(self, plan):
        """Delete the portion of a plan that has yet to occur.

        :arg plan: integer ID of a plan, as given by ``with self.plan() as plan:``

        """
        branch, turn, tick = self._btt()
        to_delete = []
        plan_ticks = self._plan_ticks[plan]
        for trn, tcks in plan_ticks.items():  # might improve performance to use a WindowDict for plan_ticks
            if turn == trn:
                for tck in tcks:
                    if tck >= tick:
                        to_delete.append((trn, tck))
            elif trn > turn:
                to_delete.extend((trn, tck) for tck in tcks)
        # Delete stuff that happened at contradicted times, and then delete the times from the plan
        where_cached = self._where_cached
        time_plan = self._time_plan
        for trn, tck in to_delete:
            for cache in where_cached[branch, trn, tck]:
                cache.remove(branch, trn, tck)
                if hasattr(cache, 'deldb'):
                    cache.deldb(branch, trn, tck)
            del where_cached[branch, trn, tck]
            plan_ticks[trn].remove(tck)
            if not plan_ticks[trn]:
                del plan_ticks[trn]
            del time_plan[branch, trn, tck]

    # easier to override things this way
    @property
    def branch(self):
        return self._get_branch()

    @branch.setter
    def branch(self, v):
        self._set_branch(v)

    def _get_turn(self):
        return self._oturn

    def _set_turn(self, v):
        if v == self.turn:
            self._otick = self._turn_end_plan[tuple(self.time)]
            return
        if not isinstance(v, int):
            raise TypeError("turn must be an integer")
        # enforce the arrow of time, if it's in effect
        if self._forward and v < self._oturn:
            raise ValueError("Can't time travel backward in a forward context")
        # first make sure the cursor is not before the start of this branch
        branch = self.branch
        tick = self._turn_end_plan.setdefault((branch, v), 0)
        parent, turn_start, tick_start, turn_end, tick_end = self._branches[branch]
        if branch != 'trunk':
            if v < turn_start:
                raise ValueError(
                    "The turn number {} "
                    "occurs before the start of "
                    "the branch {}".format(v, branch)
                )
        self._otick = tick
        self._oturn = v

    # easier to override things this way
    @property
    def turn(self):
        return self._get_turn()

    @turn.setter
    def turn(self, v):
        self._set_turn(v)

    def _get_tick(self):
        return self._otick

    def _set_tick(self, v):
        if not isinstance(v, int):
            raise TypeError("tick must be an integer")
        time = branch, turn = self._obranch, self._oturn
        # enforce the arrow of time, if it's in effect
        if self._forward and v < self._otick:
            raise ValueError("Can't time travel backward in a forward context")
        if v > self._turn_end_plan[time]:
            self._turn_end_plan[time] = v
        if not self._planning:
            if v > self._turn_end[time]:
                self._turn_end[time] = v
            parent, turn_start, tick_start, turn_end, tick_end = self._branches[branch]
            if turn == turn_end and v > tick_end:
                self._branches[branch] = parent, turn_start, tick_start, turn, v
        self._otick = v

    # easier to override things this way
    @property
    def tick(self):
        return self._get_tick()

    @tick.setter
    def tick(self, v):
        self._set_tick(v)

    def _btt(self):
        """Return the branch, turn, and tick."""
        return self._obranch, self._oturn, self._otick

    def _nbtt(self):
        """Increment the tick and return branch, turn, tick

        Unless we're viewing the past, in which case raise HistoryError.

        Idea is you use this when you want to advance time, which you
        can only do once per branch, turn, tick.

        """
        from .cache import HistoryError
        branch, turn, tick = self._btt()
        tick += 1
        if (branch, turn) in self._turn_end_plan:
            if tick > self._turn_end_plan[branch, turn]:
                self._turn_end_plan[branch, turn] = tick
            else:
                tick = self._turn_end_plan[branch, turn] + 1
        self._turn_end_plan[branch, turn] = tick
        if self._turn_end[branch, turn] > tick:
            raise HistoryError(
                "You're not at the end of turn {}. Go to tick {} to change things".format(
                    turn, self._turn_end[branch, turn]
                )
            )
        parent, turn_start, tick_start, turn_end, tick_end = self._branches[branch]
        if turn < turn_end or (
            turn == turn_end and tick < tick_end
        ):
            raise HistoryError(
                "You're in the past. Go to turn {}, tick {} to change things".format(turn_end, tick_end)
            )
        if self._planning:
            if (turn, tick) in self._plan_ticks[self._last_plan]:
                raise HistoryError(
                    "Trying to make a plan at {}, but that time already happened".format((branch, turn, tick))
                )
            self._plan_ticks[self._last_plan][turn].append(tick)
            self._plan_ticks_uncommitted.append((self._last_plan, turn, tick))
            self._time_plan[branch, turn, tick] = self._last_plan
        self._otick = tick
        return branch, turn, tick

    def commit(self):
        """Write the state of all graphs to the database and commit the transaction.

        Also saves the current branch, turn, and tick.

        """
        self.query.globl['branch'] = self._obranch
        self.query.globl['turn'] = self._oturn
        self.query.globl['tick'] = self._otick
        set_branch = self.query.set_branch
        for branch, (parent, turn_start, tick_start, turn_end, tick_end) in self._branches.items():
            set_branch(branch, parent, turn_start, tick_start, turn_end, tick_end)
        turn_end = self._turn_end
        set_turn = self.query.set_turn
        for (branch, turn), plan_end_tick in self._turn_end_plan.items():
            set_turn(branch, turn, turn_end[branch], plan_end_tick)
        if self._plans_uncommitted:
            self.query.plans_insert_many(self._plans_uncommitted)
        if self._plan_ticks_uncommitted:
            self.query.plan_ticks_insert_many(self._plan_ticks_uncommitted)
        self.query.commit()
        self._plans_uncommitted = []
        self._plan_ticks_uncommitted = []

    def close(self):
        """Write changes to database and close the connection"""
        self.commit()
        self.query.close()

    def _init_graph(self, name, type_s='Graph'):
        if self.query.have_graph(name):
            raise GraphNameError("Already have a graph by that name")
        if name in self.illegal_graph_names:
            raise GraphNameError("Illegal name")
        self.query.new_graph(name, type_s)

    def new_graph(self, name, data=None, **attr):
        """Return a new instance of type Graph, initialized with the given
        data if provided.

        :arg name: a name for the graph
        :arg data: dictionary or NetworkX graph object providing initial state

        """
        self._init_graph(name, 'Graph')
        g = Graph(self, name, data, **attr)
        self._graph_objs[name] = g
        return g

    def new_digraph(self, name, data=None, **attr):
        """Return a new instance of type DiGraph, initialized with the given
        data if provided.

        :arg name: a name for the graph
        :arg data: dictionary or NetworkX graph object providing initial state

        """
        self._init_graph(name, 'DiGraph')
        dg = DiGraph(self, name, data, **attr)
        self._graph_objs[name] = dg
        return dg

    def new_multigraph(self, name, data=None, **attr):
        """Return a new instance of type MultiGraph, initialized with the given
        data if provided.

        :arg name: a name for the graph
        :arg data: dictionary or NetworkX graph object providing initial state

        """
        self._init_graph(name, 'MultiGraph')
        mg = MultiGraph(self, name, data, **attr)
        self._graph_objs[name] = mg
        return mg

    def new_multidigraph(self, name, data=None, **attr):
        """Return a new instance of type MultiDiGraph, initialized with the given
        data if provided.

        :arg name: a name for the graph
        :arg data: dictionary or NetworkX graph object providing initial state

        """
        self._init_graph(name, 'MultiDiGraph')
        mdg = MultiDiGraph(self, name, data, **attr)
        self._graph_objs[name] = mdg
        return mdg

    def get_graph(self, name):
        """Return a graph previously created with ``new_graph``,
        ``new_digraph``, ``new_multigraph``, or
        ``new_multidigraph``

        :arg name: name of an existing graph

        """
        if name in self._graph_objs:
            return self._graph_objs[name]
        graphtypes = {
            'Graph': Graph,
            'DiGraph': DiGraph,
            'MultiGraph': MultiGraph,
            'MultiDiGraph': MultiDiGraph
        }
        type_s = self.query.graph_type(name)
        if type_s not in graphtypes:
            raise GraphNameError(
                "I don't know of a graph named {}".format(name)
            )
        g = graphtypes[type_s](self, name)
        self._graph_objs[name] = g
        return g

    def del_graph(self, name):
        """Remove all traces of a graph's existence from the database

        :arg name: name of an existing graph

        """
        # make sure the graph exists before deleting anything
        self.get_graph(name)
        self.query.del_graph(name)
        if name in self._graph_objs:
            del self._graph_objs[name]

    def _iter_parent_btt(self, branch=None, turn=None, tick=None, *, stoptime=None):
        """Private use. Iterate over (branch, turn, tick), where the branch is
        a descendant of the previous (starting with whatever branch is
        presently active and ending at 'trunk'), and the turn is the
        latest revision in the branch that matters.

        :arg stoptime: This may be a branch, in which case iteration will stop
        instead of proceeding into that branch's parent; or it may be a triple,
        ``(branch, turn, tick)``, in which case iteration will stop instead of
        yielding any time before that. The tick may be ``None``, in which case
        iteration will stop instead of yielding the turn.

        """
        branch = branch or self.branch
        trn = self.turn if turn is None else turn
        tck = self.tick if tick is None else tick
        yield branch, trn, tck
        stopbranches = set()
        if stoptime:
            if type(stoptime) is tuple:
                stopbranch = stoptime[0]
                stopbranches.add(stopbranch)
                stopbranches.update(self._branch_parents[stopbranch])
            else:
                stopbranch = stoptime
                stopbranches = self._branch_parents[stopbranch]
        _branches = self._branches
        while branch in _branches:
            # ``par`` is the parent branch;
            # ``(trn, tck)`` is when ``branch`` forked off from ``par``
            (branch, trn, tck, _, _) = _branches[branch]
            if branch in stopbranches and (
                trn < stoptime[1] or (
                    trn == stoptime[1] and (
                        stoptime[2] is None or tck <= stoptime[2]
                    )
                )
            ):
                return
            yield branch, trn, tck

    def _branch_descendants(self, branch=None):
        """Iterate over all branches immediately descended from the current
        one (or the given one, if available).

        """
        branch = branch or self.branch
        for (parent, (child, _, _, _, _)) in self._branches.items():
            if parent == branch:
                yield child

    def _node_exists(self, character, node):
        return self._nodes_cache.contains_entity(character, node, *self._btt())

    def _exist_node(self, character, node, exist=True):
        branch, turn, tick = self._nbtt()
        self.query.exist_node(
            character,
            node,
            branch,
            turn,
            tick,
            True
        )
        self._nodes_cache.store(character, node, branch, turn, tick, exist)

    def _edge_exists(self, character, orig, dest, idx=0):
        return self._edges_cache.contains_entity(
            character, orig, dest, idx, *self._btt()
        )

    def _exist_edge(
            self, character, orig, dest, idx=0, exist=True
    ):
        branch, turn, tick = self._nbtt()
        self.query.exist_edge(
            character,
            orig,
            dest,
            idx,
            branch,
            turn,
            tick,
            exist
        )
        self._edges_cache.store(
            character, orig, dest, idx, branch, turn, tick, exist
        )
